from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS
from apps.TA.storages.abstract.indicator import IndicatorStorage
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class RocStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class RocSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if str(self.index) is not "close_price":
            logger.debug(f'index {self.index} is not close_price ...ignoring...')
            return

        new_roc_storage = RocStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        for horizon in HORIZONS:
            periods = horizon * 10

            close_value_np_array = new_roc_storage.get_denoted_price_array("close_price", periods)

            timeperiod = min([len(close_value_np_array), periods])
            roc_value = talib.ROC(close_value_np_array, timeperiod=timeperiod)[-1]
            # logger.debug(f'savingRoc value {roc_value} for {self.ticker} on {periods} periods')

            new_roc_storage.periods = periods
            new_roc_storage.value = float(roc_value)
            new_roc_storage.save()



import os
import logging
import redis
# from redisgraph import Graph

logger = logging.getLogger('POPOTO-REDIS_DB')

global POPOTO_REDIS_DB
# global REDIS_GRAPH
BEGINNING_OF_TIME = 0
ENCODING = 'utf-8'

try:
    BEGINNING_OF_TIME = int(os.environ.get("BEGINNING_OF_TIME", 0))
except ValueError:
    logger.critical("BEGINNING_OF_TIME is set but should be an integer in unix time seconds where 0 equals 1970-01-01")
except Exception as e:
    logger.debug(e)

try:
    REDIS_URL = os.environ.get("REDIS_URL", "")
    if REDIS_URL:
        POPOTO_REDIS_DB = redis.from_url(REDIS_URL)
        logger.debug("Redis connection established.")
    else:
        REDIS_HOST, REDIS_PORT = "127.0.0.1:6379".split(":")
        pool = redis.ConnectionPool(host=REDIS_HOST, port=REDIS_PORT, db=0)
        POPOTO_REDIS_DB = redis.Redis(connection_pool=pool)
        # REDIS_GRAPH = Graph('social', POPOTO_REDIS_DB)

except Exception as e:
    logger.info(str(e))


def set_REDIS_DB_settings(env_partition_name: str = "", *args, **kwargs):
    # todo: use this to mark keys in redis db, so they can be separated and deleted
    env_partition_name = env_partition_name or os.environ.get('ENV', "")

    global POPOTO_REDIS_DB
    POPOTO_REDIS_DB = redis.Redis(*args, **kwargs)
    # global REDIS_GRAPH
    # REDIS_GRAPH = Graph('social', POPOTO_REDIS_DB)
    logger.debug("Redis connection reset.")


def get_REDIS_DB():
    return POPOTO_REDIS_DB


def print_redis_info():
    logger.info(POPOTO_REDIS_DB.info())

    used_memory, maxmemory = int(POPOTO_REDIS_DB.info()['used_memory']), int(POPOTO_REDIS_DB.info()['maxmemory'])
    maxmemory_human = POPOTO_REDIS_DB.info()['maxmemory_human']
    if maxmemory and maxmemory > 0:
        logger.info(f"Redis currently consumes {round(100 * used_memory / maxmemory, 2)}% out of {maxmemory_human}")


class PopotoException(Exception):
    def __init__(self, message):
        self.message = message
        logger.error(message)



import logging
from src.popoto.exceptions import FinanceException
from src.popoto._archive.timeseries import TimeseriesModel

logger = logging.getLogger(__name__)


class TickerException(FinanceException):
    pass


class TickerStorage(TimeseriesModel):
    """
    stores timeseries data on tickers in a sorted set unique to each ticker and publisher (data source)
    """
    class_describer = "ticker"
    value_sig_figs = 6

    def __init__(self, *args, **kwargs):
        """
        'ticker' REQUIRED
        'publisher EXPECTED BUT CAN STILL SAVE WITHOUT
        """
        super().__init__(*args, **kwargs)
        try:
            self.ticker = str(kwargs['ticker'])  # str eg. BTC_USD
            self.publisher = str(kwargs['publisher'])  # str eg. binance
        except KeyError:
            raise Exception("Indicator requires a ticker and publisher as parameters")
        except Exception as e:
            raise Exception(str(e))

        if self.ticker.find("_") <= 0:
            raise Exception("ticker should be like GOOG_USD or LTC_BTC")

        self._db_key_prefix = f'{self.ticker}:{self.publisher}:'
        self._db_key = self.build_db_key()


    @classmethod
    def query(cls, *args, **kwargs):

        ticker = kwargs.get("ticker", None)
        if not ticker:
            raise TickerException("ticker required for ticker query")
        publisher = kwargs.get("publisher", "")
        kwargs["key_prefix"] = f'{ticker}:{publisher}'

        results_dict = super().query(*args, **kwargs)
        if results_dict:
            results_dict['publisher'] = publisher
            results_dict['ticker'] = ticker
        return results_dict

    def save(self, publish=False, pipeline=None, *args, **kwargs):
        self.value = '{:g}'.format(float('{:.{p}g}'.format(self.value, p=self.value_sig_figs)))
        return super().save(publish=publish, pipeline=pipeline, *args, **kwargs)



import time
from abc import ABC
import pandas as pd
from ...exceptions import FinanceException
from ...models import TimeseriesModel
from ...finance import PriceStorage, VolumeStorage

# get_nearest_1day_score
# PERIODS_24HR, DEFAULT_PRICE_INDEXES, DEFAULT_VOLUME_INDEXES


class OLHCVException(FinanceException):
    pass


class OLHCVModel(ABC):  # candles data for an asset-symbol
    def __init__(self, symbol: str, publisher: str = "other", timestamp: int = 0, days_range: int = 1, period="daily"):
        self.symbol = symbol
        self.dataframe = pd.DataFrame()
        self.publisher = publisher
        self.timestamp = timestamp or int(time.time())
        self.period = period
        self.days_range = days_range


    def get_dataframe(self, timestamp: int = 0, days_range: int = 0):
        self.timestamp = timestamp or self.timestamp
        self.days_range = days_range or self.days_range

        dataframes = {}
        low_score = high_score = TimeseriesModel.score_from_timestamp(self.timestamp)

        for index in reversed(DEFAULT_PRICE_INDEXES):
            price_timeseries = PriceStorage.query(
                ticker=f"{self.symbol}_USD",
                index=index,
                publisher=self.publisher,
                timestamp=self.timestamp,
                periods_range=float(self.days_range) * PERIODS_24HR * 1.1  # n days x 24 hours * 110%
            )
            if 'scores' not in price_timeseries or int(price_timeseries['values_count']) < 1:
                continue

            dataframes[index] = pd.DataFrame(
                index=[get_nearest_1day_score(score) for score in price_timeseries['scores']],
                data={index: price_timeseries['values'],},
                dtype=float
            )
            low_score = min([low_score, price_timeseries['scores'][0]])
            high_score = max([high_score, price_timeseries['scores'][-1]])

        for index in DEFAULT_VOLUME_INDEXES:
            volume_timeseries = VolumeStorage.query(
                ticker=f"{self.symbol}_USD",
                index="close_volume",
                publisher=self.publisher,
                timestamp=self.timestamp,
                timestamp_tolerance=12 * 12,  # 12 hrs
                periods_range=float(self.days_range) * PERIODS_24HR * 1.1  # n days x 24 hours * 110%
            )
            if 'scores' in volume_timeseries:
                dataframes[index] = pd.DataFrame(
                    index=[get_nearest_1day_score(score) for score in volume_timeseries['scores']],
                    data={index: volume_timeseries['values'],},
                    dtype=float
                )
                low_score = min([low_score, volume_timeseries['scores'][0]])
                high_score = max([high_score, volume_timeseries['scores'][-1]])


        # new empty dataframe with every daily score included in the index
        self.dataframe = pd.DataFrame(
            index=range(get_nearest_1day_score(low_score),
                        get_nearest_1day_score(high_score),
                        PERIODS_24HR)
        )

        # merge all dataframes
        for index in reversed(DEFAULT_PRICE_INDEXES):  # close_price goes first
            self.dataframe = pd.merge(
                self.dataframe, dataframes[index],
                left_index=True, right_index=True,
                how="outer"
            )
        for index in DEFAULT_VOLUME_INDEXES:
            self.dataframe = pd.merge(
                self.dataframe, dataframes[index],
                left_index=True, right_index=True,
                how="outer"
            )

        # fill-forward missing close_prices (includes weekends for stocks)
        self.dataframe['close_price'].fillna(method="ffill", axis=0, inplace=True)
        for index in reversed(DEFAULT_PRICE_INDEXES):  # close_price goes first
            # fill all other missing prices with close price
            self.dataframe[index].fillna(self.dataframe['close_price'], axis=0, inplace=True)

        # fill missing volumes with 0
        for index in DEFAULT_VOLUME_INDEXES:
            self.dataframe[index].fillna(0, axis=0, inplace=True)

        # self.dataframe = self.dataframe.astype({'score': 'int64'})
        self.dataframe.index.name = "score"
        self.dataframe = self.dataframe.reset_index().drop_duplicates(subset='score', keep='last').set_index('score').sort_index()
        return self.dataframe




import logging


from src.popoto.exceptions import FinanceException
from src.popoto.finance import TickerStorage

# HORIZONS

logger = logging.getLogger(__name__)

TRENDS = (BEARISH, BULLISH, OTHER) = (-1, 1, 0)


class IndicatorException(FinanceException):
    pass


class IndicatorStorage(TickerStorage):
    """
    stores indicators in a sorted set unique to each ticker and publisher
    requires data to be a resampling to represent the most recent 5min block of time
    timestamp value must be evenly divisible by 5 minutes (300 seconds)
    add short, medium, long as 1hr, 4hr, 24hr time horizons
    """
    class_describer = "indicator"
    value_sig_figs = 6

    class_periods_list = [1,]  # class should override this
    # list of integers where for x: (1 <= x <= 200)

    requisite_pv_indexes = []  # class should override this.

    # may only include values in default_price_indexes or default_volume_indexes
    # eg. ["high_price", "low_price", "open_price", "close_price", "close_volume"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # ALL INDICATORS ARE ASSUMED 5-MIN PERIOD RESAMPLED
        if self.unix_timestamp % 3600 != 0:
            raise IndicatorException("indicator timestamp should be % 3600")

        # self.horizon = int(kwargs.get('horizon', 1))
        self.periods = int(kwargs.get('periods', 1))  # * self.horizon))

        # if self.periods // self.horizon == 0:
        #     raise IndicatorException(f'horizon {self.horizon} '
        #                              f'must be less than periods {self.periods} (or ==1)')
        # elif self.periods % self.horizon != 0:
        #     raise IndicatorException(f'horizon {self.horizon} '
        #                              f'must be a factor of periods {self.periods}')

        self.db_key_suffix = f':{self.periods}'
        self.value = None

    def get_value(self, refresh_from_db=False):
        try:
            if self.value and not refresh_from_db:
                pass
            else:
                self.value = self.query(
                    ticker=self.ticker, publisher=self.publisher, timestamp=self.unix_timestamp
                )['values'][-1]
            if not self.value:
                self.value = self.compute_value()
        except IndexError:
            self.value = None  # value not found
        except Exception as e:
            logger.error(str(e))
            self.value = None

        return self.value

    @classmethod
    def score_from_timestamp(cls, *args, **kwargs) -> int:
        # enforce integer score == num of 5min periods since Jan_1_2017
        return int(round(super().score_from_timestamp(*args, **kwargs)))

    @classmethod
    def periods_from_seconds(cls, *args, **kwargs) -> int:
        # enforce integer periods, round to nearest
        return int(round(super().periods_from_seconds(*args, **kwargs)))

    @classmethod
    def query(cls, *args, **kwargs):

        periods_key = kwargs.get("periods_key", "")
        key_suffix = kwargs.get("key_suffix", "")

        if periods_key:
            kwargs["key_suffix"] = f'{periods_key}' + (f':{key_suffix}' if key_suffix else "")

        results_dict = super().query(*args, **kwargs)

        results_dict['periods_key'] = periods_key
        return results_dict

    @classmethod
    def get_periods_list(cls):
        periods_list = []
        for s in cls.class_periods_list:
            periods_list.extend([h * s for h in HORIZONS])
        return set(periods_list)

    def get_denoted_price_array(self, index: str = "close_price", periods: int = 0):
        from apps.TA.storages.data.price import PriceStorage
        results_dict = PriceStorage.query(
            ticker=self.ticker,
            publisher=self.publisher,
            index=index,
            timestamp=self.unix_timestamp,
            periods_range=periods or self.periods
        )
        return self.get_values_array_from_query(results_dict, limit=periods)

    def compute_value(self, periods: int = 0) -> str:
        periods = periods or self.periods

        index_value_arrrays = {}
        for index in self.requisite_pv_indexes:
            index_value_arrrays[index] = self.get_denoted_price_array(index, periods)
            if not len(index_value_arrrays[index]): return ""

        return self.compute_value_with_requisite_indexes(index_value_arrrays, periods)

    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        """
        custom class should set cls.requisite_pv_indexes
        override this function with custom logic

        :param index_value_arrrays: a dict with keys matching requisite+pv_indexes and values from self.get_denoted_price_array()
        :param periods: number of periods to compute value for
        :return:
        """
        # example:
        # import talib, math
        # periods = periods or self.periods
        # sma_value = talib.SMA(requisite_pv_index_arrrays["close_price"], timeperiod=periods)[-1]
        # if math.isnan(sma_value):
        #     return ""
        # return str(sma_value)
        return ""

    def compute_and_save(self) -> bool:
        """

        :return: True if value saved, else False
        """

        if not all([
            self.ticker, self.publisher, self.unix_timestamp, self.periods
        ]):
            raise Exception("missing required values")

        self.value = self.compute_value(self.periods)
        if self.value:
            self.save()
        return bool(self.value)

    @classmethod
    def compute_and_save_all_values_for_timestamp(cls, ticker, publisher, timestamp):
        new_class_storage = cls(ticker=ticker, publisher=publisher, timestamp=timestamp)
        for periods in cls.get_periods_list():
            new_class_storage.periods = periods
            new_class_storage.compute_and_save()

    def produce_signal(self):
        """
        overwrite me, defining the criteria for sending signals

        :return: None
        """
        if "this indicator" == "interesting":
            self.send_signal(trend=BULLISH)

    def send_signal(self, trend=OTHER, *args, **kwargs):
        """
        :param trend: BULLISH, BEARISH, or OTHER
        :param args:
        :param kwargs:
            add these optional kwargs:
            strength_value = 1,
            strength_max = 5,
        :return: signal object (Django model object)
        """
        from src.popoto.finance import PriceStorage
        price_results_dict = PriceStorage.query(ticker=self.ticker, publisher=self.publisher)
        most_recent_price = int(price_results_dict['values'][0])
        # from apps.TA.storages.data.volume import VolumeStorage
        # volume_results_dict = VolumeStorage.query(ticker=self.ticker, publisher=self.publisher)
        # most_recent_volume = float(volume_results_dict ['values'][0])


    def save(self, *args, **kwargs):

        # check meets basic requirements for saving
        if not all([self.ticker, self.publisher,
                    self.periods, self.value,
                    self.unix_timestamp]):
            logger.error("incomplete information, cannot save \n" + str(self.__dict__))
            raise IndicatorException("save error, missing data")

        self.db_key_suffix = f'{str(self.periods)}'
        save_result = super().save(*args, **kwargs)
        try:
            self.produce_signal()
        except Exception as e:
            logger.error("error producing signal for indicator" + str(e))
        return save_result


"""
===== EXAMPLE USAGE =====

my_indicator = TimeseriesIndicator(ticker="ETH_BTC",
                                   publisher="bittrex",
                                   timestamp=1483228800,
                                   periods=12*20)
my_indicator.value = "BUY BITCOIN"
my_indicator.save()

# advanced:

from settings.redis_db import database
pipeline = database.pipeline()
for thing in ['towel', 42, 'babelfish', 'vogon poetry']:
    very_special_signal.unix_timestamp += 300
    very_special_signal.value = thing
    pipeline = very_special_signal.save(pipeline)
pipeline.execute()

===== EXAMPLE USAGE =====
"""



from datetime import date, datetime, time
from decimal import Decimal
import redis

from ..exceptions import ModelException

import logging

from ..models.db_key import DB_key

logger = logging.getLogger('POPOTO.field')

VALID_FIELD_TYPES = {
    int, float, Decimal, str, bool, bytes, list, dict, set, tuple, date, datetime, time,
}


class FieldBase(type):
    """Metaclass for all Popoto Fields."""

    def __new__(cls, name, bases, attrs, **kwargs):
        # if not a Field, skip setup
        parents = [b for b in bases if isinstance(b, FieldBase)]
        if not parents:
            return super().__new__(cls, name, bases, attrs, **kwargs)

        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        new_class.field_class_key = DB_key(f"${name.strip('Field')}F")
        return new_class


class Field(metaclass=FieldBase):
    type: type = str
    key: bool = False
    unique: bool = False
    auto: bool = False
    null: bool = True
    value: str = None
    max_length: int = 1024
    default = ""
    sorted: bool = False

    def __init__(self, **kwargs):
        self.field_defaults = {  # default
            'type': str,
            'key': False,
            'unique': False,
            'auto': False,
            'null': True,
            'value': None,
            'max_length': 1024,  # Redis limit is 512MB
            'default': None,
            'sorted': False,
        }
        # set field_options, let kwargs override
        field_options = self.field_defaults.copy()
        for k, v in field_options.items():
            setattr(self, k, kwargs.get(k, v))
        if self.__class__ == Field and self.type not in VALID_FIELD_TYPES:
            raise ModelException(f"{self.type} is not a valid Field type")

    @classmethod
    def is_valid(cls, field, value, null_check=True, **kwargs) -> bool:
        if not null_check and value is None:
            return True
        if field.null and value is None:
            return True
        elif value is None:
            logger.error(f"field {field} is null")
            return False
        elif not isinstance(value, field.type):
            logger.error(f"field {field} is type {field.type}. But value is {type(value)}")
            return False
        if field.type == str and len(str(value)) > field.max_length:
            logger.error(f"{field} value is greater than max_length={field.max_length}")
            return False
        return True

    def format_value_pre_save(self, field_value):
        """
        format field_value before saving to db
        return corrected field_value
        assumes validation is already passed
        """
        return field_value

    @classmethod
    def get_special_use_field_db_key(cls, model: 'Model', field_name: str) -> DB_key:
        """
        For use by child class when implementing additional Redis data structures
        Children implementing more than one new structure will need to augment this.
        """
        return DB_key(cls.field_class_key, model._meta.db_class_key, field_name)


    @classmethod
    def on_save(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None,
                **kwargs):
        """
        for parent classes to override.
        will run for every field of the model instance, including null attributes
        runs async with model instance save event, so order of processing is not guaranteed
        """
        # todo: create choice Sets of instance keys for fields using choices option
        # if model_instance._meta.fields[field_name].choices:
        #     # this will not work! how to edit, delete, prevent overwrite and duplicates?
        #     field_value_b = cls.encode(field_value)
        #     if pipeline:
        #         return pipeline.set(cls.get_special_use_field_db_key(model_instance, field_name), field_value_b)
        #     else:
        #         from ..redis_db import POPOTO_REDIS_DB
        #         return POPOTO_REDIS_DB.set(cls.get_special_use_field_db_key(model_instance, field_name), field_value_b)

        return pipeline if pipeline else None

    @classmethod
    def on_delete(cls, model_instance: 'Model', field_name: str, field_value, pipeline=None, **kwargs):
        """
        for parent classes to override.
        will run for every field of the model instance, including null attributes
        runs async with model instance delete event, so order of processing is not guaranteed
        """
        return pipeline if pipeline else None

    def get_filter_query_params(self, field_name: str) -> list:
        # todo: refactor this method to return a set, not a list
        if self.key:
            params = [f'{field_name}', ]
            if self.null:
                params += [f'{field_name}__isnull', ]
            return params
        return list()

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """
        from ..models.query import QueryException
        raise QueryException("Query filter not allowed on base Field. Consider using a KeyField")



# https://stackoverflow.com/questions/16974047/efficient-way-to-find-missing-elements-in-an-integer-sequence/16974075#16974075
from itertools import islice, chain

def window(seq, n=2):
    "Returns a sliding window (of width n) over data from the iterable"
    "   s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def missing_elements(L):
    missing = chain.from_iterable(range(x + 1, y) for x, y in window(L) if (y - x) > 1)
    return list(missing)



from abc import ABC
import logging
from ..redis_db import POPOTO_REDIS_DB, ENCODING
import msgpack

logger = logging.getLogger('POPOTO-subscriber')


class SubscriberException(Exception):
    pass


class Subscriber(ABC):
    sub_channel_names: list = []

    def __init__(self, *args, **kwargs):
        self.pubsub = POPOTO_REDIS_DB.pubsub()
        logger.info(f'New pubsub for {self.__class__.__name__}')
        for channel_name in self.sub_channel_names:
            self.pubsub.subscribe(channel_name)
            logger.info(f'{self.__class__.__name__} subscribed to {channel_name} channel')

    def __call__(self):
        import msgpack_numpy as m
        m.patch()
        data_event = self.pubsub.get_message()
        if not data_event:
            return
        if not data_event.get('type') == 'message':
            return

        # logger.debug(f"received message: {data_event}")

        try:
            channel_name = data_event.get('channel').decode(ENCODING)
            event_data = msgpack.unpackb(data_event.get('data'))
            logger.debug(f'handling event in {self.__class__.__name__}')
            self.pre_handle(channel_name, event_data)
            self.handle(channel_name, event_data)
        except KeyError as e:
            logger.warning(f'unexpected format: {data_event} ' + str(e))
            pass  # message not in expected format, just ignore
        except msgpack.exceptions.FormatError:
            logger.warning(f'unexpected data format: {data_event["data"]}')
            pass  # message not in expected format, just ignore
        except Exception as e:
            raise SubscriberException(f'Error calling {self.__class__.__name__}: ' + str(e))

    def pre_handle(self, channel, data, *args, **kwargs):
        pass

    def handle(self, channel, data, *args, **kwargs):
        """
        overwrite me with some logic
        :return: None
        """
        logger.warning(f'NEW MESSAGE for '
                       f'{self.__class__.__name__} subscribed to '
                       f'{channel} channel '
                       f'BUT HANDLER NOT DEFINED! '
                       f'... message/event discarded')
        pass



from apps.TA import IndicatorException, IndicatorStorage
from apps.TA.storages.abstract.ticker_subscriber import TickerSubscriber



class IndicatorSubscriber(TickerSubscriber):
    class_describer = "indicator_subscriber"
    classes_subscribing_to = [
        # ...
    ]
    storage_class = IndicatorStorage  # override with applicable storage class

    def extract_params(self, channel, data, *args, **kwargs):

        # parse data like...
        # {
        #     "key": "VEN_USDT:binance:PriceStorage:close_price",
        #     "name": "176760000:1532373300",
        #     "score": "1532373300"
        # }

        [self.ticker, self.publisher, object_class, self.key_suffix] = data["key"].split(":")

        if not object_class == channel and object_class in [
            sub_class.__name__ for sub_class in self.classes_subscribing_to
        ]:
            logger.warning(f'Unexpected that these are not the same:'
                           f'object_class: {object_class}, '
                           f'channel: {channel}, '
                           f'subscribing classes: {self.classes_subscribing_to}')

        [value, score] = data["name"].split(":")
        self.value, self.score = float(value), float(score)

        if not self.score == int(float(data["score"])):
            logger.warning(f'Unexpected that score in name {self.score} '
                           f'is different than score {data["score"]}')

        if not self.score == int(self.score):
            raise IndicatorException("indicator timestamp should be 5min timestamp, score should be whole number (int)")

        self.timestamp = IndicatorStorage.timestamp_from_score(self.score)

        return


    def pre_handle(self, channel, data, *args, **kwargs):
        self.extract_params(channel, data, *args, **kwargs)
        super().pre_handle(channel, data, *args, **kwargs)


    def handle(self, channel, data, *args, **kwargs):
        """
        a new instance was saved belonging to one of the classes subscribed to
        for standard indicators self.ticker, self.publisher, and self.timestamp are available
        these values were collected by self.pre_handle() and self.extract_params()

        :param channel:
        :param data:
        :param args:
        :param kwargs:
        :return:
        """

        if self.key_suffix not in self.storage_class.requisite_pv_indexes:
            logger.debug(f'index {self.key_suffix} is not in {self.storage_class.requisite_pv_indexes} ...ignoring...')
            return

        self.storage_class.compute_and_save_all_values_for_timestamp(self.ticker, self.publisher, self.timestamp)



from .field import Field
from .key_field_mixin import KeyFieldMixin
from .auto_field_mixin import AutoFieldMixin
from .sorted_field_mixin import SortedFieldMixin


class IntField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = int
        super().__init__(**kwargs)


class FloatField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = float
        super().__init__(**kwargs)


class DecimalField(Field):
    def __init__(self, *args, **kwargs):
        from decimal import Decimal
        kwargs['type'] = Decimal
        super().__init__(**kwargs)


class StringField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = str
        super().__init__(**kwargs)


class BooleanField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = bool
        super().__init__(**kwargs)


class BytesField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = bytes
        super().__init__(**kwargs)


class ListField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = list
        super().__init__(**kwargs)


class DictField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = dict
        super().__init__(**kwargs)


class SetField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = set
        super().__init__(**kwargs)


class TupleField(Field):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = tuple
        super().__init__(**kwargs)


class DateField(Field):
    def __init__(self, *args, **kwargs):
        from datetime import date
        kwargs['type'] = date
        super().__init__(**kwargs)


class DatetimeField(Field):
    def __init__(self, *args, **kwargs):
        from datetime import datetime
        kwargs['type'] = datetime
        super().__init__(**kwargs)


class TimeField(Field):
    def __init__(self, *args, **kwargs):
        from datetime import time
        kwargs['type'] = time
        super().__init__(**kwargs)


class KeyField(KeyFieldMixin, Field):
    def __init__(self, *args, **kwargs):
        kwargs['key'] = True
        super().__init__(**kwargs)


class UniqueKeyField(KeyField):
    def __init__(self, *args, **kwargs):
        kwargs['unique'] = True
        kwargs['null'] = False
        super().__init__(**kwargs)


class AutoKeyField(AutoFieldMixin, UniqueKeyField):
    def __init__(self, *args, **kwargs):
        kwargs['auto'] = True
        super().__init__(**kwargs)


class SortedField(SortedFieldMixin, Field):
    def __init__(self, *args, **kwargs):
        kwargs['sorted'] = True
        super().__init__(**kwargs)


class SortedKeyField(SortedFieldMixin, KeyFieldMixin, Field):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)



import logging
from ...finance.models.ticker import TickerStorage, TickerException


logger = logging.getLogger(__name__)


default_price_indexes = ["high_price", "low_price", "open_price", "close_price", ]
derived_price_indexes = ["midpoint_price", "mean_price", "price_variance",]
default_volume_indexes = ["close_volume",]
derived_volume_indexes = ["open_volume", "high_volume", "low_volume",]
default_indexes = default_price_indexes + default_volume_indexes
derived_indexes = derived_price_indexes + derived_volume_indexes
all_indexes = default_indexes + derived_indexes

class PriceVolumeHistoryException(TickerException):
    pass


class PriceVolumeHistoryStorage(TickerStorage):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = kwargs.get('index', "close_price")
        self.value = kwargs.get('value')


    @classmethod
    def query(cls, *args, **kwargs):

        # "ETH_BTC:poloniex:PriceVolumeHistoryStorage:close_price"
        # f'{ticker}:{publisher}:{cls.__name__}:{index}'

        if not "index" in kwargs:
            logger.debug("assuming to use `close_price` index in price query")
        index = kwargs["index"] = kwargs.get("index", "close_price")

        if kwargs.get("key_suffix", None):
            logger.warning("`key_suffix` has been removed from your query. "
                           "it cannot be used in query for PriceVolumeHistoryStorage. "
                           "Only an `index` can be specified and if not provided, "
                           "it defaults to 'close_price'")

        kwargs["key_suffix"] = f':{index}'

        results_dict = super().query(*args, **kwargs)

        if results_dict:
            results_dict['index'] = index
        return results_dict


    @classmethod
    def destroy(cls, *args, **kwargs):
        pass


    def save(self, *args, **kwargs):

        # meets basic requirements for saving
        if not all([self.ticker, self.publisher,
                   self.index, self.value,
                   self.unix_timestamp]):
            logger.error("incomplete information, cannot save \n" + str(self.__dict__))
            raise PriceVolumeHistoryException("save error, missing data")

        if not self.force_save:
            if not self.index in default_price_indexes + default_volume_indexes:
                logger.error("price index not in approved list, raising exception...")
                raise PriceVolumeHistoryException("unknown index: " + str(self.index))

        self.db_key_suffix = f':{self.index}'
        # logger.debug("ready to save, db_key will be " + self.get_db_key())
        return super().save(*args, **kwargs)



from settings import LOAD_TALIB
if LOAD_TALIB:
    import math, talib

from apps.TA.storages.abstract.indicator import IndicatorStorage
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger

WMA_LIST = [30, 50, 200,]


class WmaStorage(IndicatorStorage):
    # sorted_set_key = "BTC_USDT:poloniex:WmaStorage:30"

    class_periods_list = WMA_LIST
    requisite_pv_indexes = ["close_price"]

    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        """
        with cls.requisite_pv_indexes set

        :param index_value_arrrays: a dict with keys matching requisite+pv_indexes and values from self.get_denoted_price_array()
        :param periods: number of periods to compute value for
        :return:
        """

        wma_value = talib.WMA(
            requisite_pv_index_arrrays["close_price"],
            timeperiod=periods or self.periods
        )[-1]

        logger.debug(f"WMA computed: {wma_value}")

        if math.isnan(wma_value): return ""

        return str(wma_value)

    def produce_signal(self):
        pass


class WmaSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [PriceStorage]
    storage_class = WmaStorage



from decimal import Decimal
from datetime import date, datetime, time
import redis.client
import logging
from ..models.db_key import DB_key

logger = logging.getLogger('POPOTO.KeyFieldMixin')

from ..exceptions import ModelException
from ..models.query import QueryException
from ..redis_db import POPOTO_REDIS_DB

VALID_KEYFIELD_TYPES = [
    int, float, Decimal, str, bool, date, datetime, time,
]


class KeyFieldMixin:
    """
    A KeyField is for unique identifiers and category searches
    All KeyFields are used for indexing on the DB.
    All keys together have unique_together enforced.

    UniqueKeyField and AutoKeyField provide unique constraint and auto-id generation respectively.
    All models must have one or more KeyFields.
    However an AutoKeyField will be automatically added to models without any specified KeyFields.
    """
    key: bool = True
    max_length: int = 128

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        keyfield_defaults = {
            'key': True,
            'max_length': 128,  # Redis limit is 512MB
        }
        self.field_defaults.update(keyfield_defaults)
        # set field options, let kwargs override
        for k, v in keyfield_defaults.items():
            setattr(self, k, kwargs.get(k, v))
        if self.key and self.type not in VALID_KEYFIELD_TYPES:
            raise ModelException(f"{self.type} is not a valid KeyField type")

    @classmethod
    def is_valid(cls, field, value, null_check=True, **kwargs) -> bool:
        if not super().is_valid(field, value, null_check):
            return False
        return True

    @classmethod
    def on_save(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None,
                **kwargs):
        if model_instance._meta.fields[field_name].auto:
            return pipeline if pipeline else None

        unique_set_key = DB_key(cls.get_special_use_field_db_key(model_instance, field_name), field_value)
        if pipeline:
            return pipeline.sadd(unique_set_key.redis_key, model_instance.db_key.redis_key)
        else:
            return POPOTO_REDIS_DB.sadd(unique_set_key.redis_key, model_instance.db_key.redis_key)

    @classmethod
    def on_delete(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None,
                  **kwargs):
        if model_instance._meta.fields[field_name].auto:
            return pipeline if pipeline else None

        unique_set_key = DB_key(cls.get_special_use_field_db_key(model_instance, field_name), field_value)
        if pipeline:
            return pipeline.srem(unique_set_key.redis_key, model_instance.db_key.redis_key)
        else:
            return POPOTO_REDIS_DB.srem(unique_set_key.redis_key, model_instance.db_key.redis_key)

    def get_filter_query_params(self, field_name: str) -> list:
        return super().get_filter_query_params(field_name) + [
            f'{field_name}',  # takes a str, exact match :x:
            f'{field_name}__contains',  # takes a str, matches :*x*:
            f'{field_name}__startswith',  # takes a str, matches :x*:
            f'{field_name}__endswith',  # takes a str, matches :*x:
            f'{field_name}__in',  # takes a list, returns any matches
        ]

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """

        keys_lists_to_intersect = list()
        db_key_length, field_key_position = model._meta.db_key_length, model._meta.get_db_key_index_position(field_name)
        num_keys_before = field_key_position
        num_keys_after = db_key_length - field_key_position - 1

        pipeline = POPOTO_REDIS_DB.pipeline()

        def get_key_pattern(query_value_pattern):
            key_pattern = model._meta.db_class_key.redis_key + ":"
            key_pattern += "*:" * (num_keys_before - 1)
            key_pattern += query_value_pattern
            key_pattern += ":*" * num_keys_after
            return key_pattern

        redis_set_key_prefix = model._meta.fields[field_name].get_special_use_field_db_key(model, field_name)

        for query_param, query_value in query_params.items():

            if query_param.endswith('__in'):
                pipeline_2 = POPOTO_REDIS_DB.pipeline()
                for query_value_elem in query_value:
                    pipeline_2 = pipeline_2.smembers(DB_key(redis_set_key_prefix, query_value_elem).redis_key)
                keys_lists_to_union = pipeline_2.execute()
                keys_lists_to_intersect.append(set.union(*[set(key_list) for key_list in keys_lists_to_union]))

            else:
                if query_param == f'{field_name}':
                    if model._meta.fields[field_name].auto:
                        # todo: refactor or show warning or depricate ability
                        keys_lists_to_intersect.append(
                            POPOTO_REDIS_DB.keys(get_key_pattern(query_value))
                        )
                    else:
                        keys_lists_to_intersect.append(
                            POPOTO_REDIS_DB.smembers(DB_key(redis_set_key_prefix, query_value).redis_key)
                        )

                elif query_param.endswith('__isnull'):
                    if query_value is True:
                        keys_lists_to_intersect.append(
                            POPOTO_REDIS_DB.smembers(DB_key(redis_set_key_prefix, None).redis_key)
                        )
                    elif query_value is False:
                        pipeline = pipeline.keys(get_key_pattern(f"[^None]"))  # todo: refactor
                    else:
                        raise QueryException(f"{query_param} filter must be True or False")

                elif query_param.endswith('__startswith'):
                    pipeline = pipeline.keys(get_key_pattern(f"{DB_key.clean(query_value)}*"))  # todo: refactor
                elif query_param.endswith('__endswith'):
                    pipeline = pipeline.keys(get_key_pattern(f"*{DB_key.clean(query_value)}"))  # todo: refactor

            # todo: refactor to use HSCAN or sets, https://redis.io/commands/keys
            # https://redis-py.readthedocs.io/en/stable/index.html#redis.Redis.hscan_iter

        keys_lists_to_intersect += pipeline.execute()
        logger.debug(keys_lists_to_intersect)
        if len(keys_lists_to_intersect):
            return set.intersection(*[set(key_list) for key_list in keys_lists_to_intersect])
        return set()



import logging
import uuid

import redis

logger = logging.getLogger('POPOTO.field')

class AutoFieldMixin:
    """
    AutoKeyField() is equivalent to KeyField(unique-True, auto=True)
    The AutoKeyField is an auto-generated, universally unique key
    It will be automatically added to models with no specified KeyFields
    Include this field in your model if you cannot otherwise enforce a unique-together constraint with other KeyFields.
    They auto-generated key is random and newly generated for a model instance.
    Model instances with otherwise identical properties are saved as separate instances with different auto-keys.
    """
    # todo: add support for https://github.com/ai/nanoid

    auto: bool = True
    auto_uuid_length: int = 32
    auto_id: str = ""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        autokeyfield_defaults = {
            'auto': True,
            'auto_uuid_length': 32,
            'auto_id': "",
        }
        self.field_defaults.update(autokeyfield_defaults)
        # set field options, let kwargs override
        for k, v in autokeyfield_defaults.items():
            setattr(self, k, kwargs.get(k, v))

    @classmethod
    def is_valid(cls, field, value, null_check=True, **kwargs) -> bool:
        if value and len(value) != field.auto_uuid_length:
            logger.error(f"auto key value is length {len(value)}. It should be {field.auto_uuid_length}")
            return False
        return super().is_valid(field, value, null_check=null_check, **kwargs)

    def get_new_auto_key_value(self):
        return uuid.uuid4().hex[:self.auto_uuid_length]

    def set_auto_key_value(self, force: bool = False):
        if self.auto or force:
            self.default = self.get_new_auto_key_value()

    @classmethod
    def on_save(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None, **kwargs):
        return pipeline if pipeline else None





from settings import LOAD_TALIB

if LOAD_TALIB:
    import math, talib

from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH, BEARISH, OTHER
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger

EXAMPLE_INDICATOR_PERIODS_LIST = [20, 30, 50]


class ExampleIndicatorStorage(IndicatorStorage):
    # example sorted_set_key = "BTC_USDT:poloniex:ExampleIndicatorStorage:20"

    class_periods_list = EXAMPLE_INDICATOR_PERIODS_LIST
    requisite_pv_indexes = ["close_price", "high_price"]

    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        """
        with cls.requisite_pv_indexes set

        :param index_value_arrrays: a dict with keys matching requisite+pv_indexes and values from self.get_denoted_price_array()
        :param periods: number of periods to compute value for
        :return:
        """
        periods = periods or self.periods  # best practice to keep this line

        example_indicator_value = talib.EXAMPLE(
            requisite_pv_index_arrrays["close_price"],  # 'close_price' is set as a requirement on line 17
            requisite_pv_index_arrrays["high_price"],  # 'close_price' is set as a requirement on line 17
            timeperiod=periods  # usually you are sampling the indicator on the same set of periods as the pv data
        )[-1]  # talib will return an array, so just take the last value

        if math.isnan(example_indicator_value):
            return ""  # best practice to return empty string if no value is available

        return str(example_indicator_value)  # this will save to self.value for each instance
        # for tuple values, format as a string first: eg. f"{example_indicator_value[0]}:{example_indicator_value[1]}"

    def produce_signal(self):
        """
        defining the criteria for sending signals

        :return: None
        """
        if self.value == "interesting":
            self.send_signal(trend=BULLISH)


class ExampleIndicatorSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage  # the handle function below will be called each time a new instance of this class is saved
    ]
    storage_class = ExampleIndicatorStorage

    def handle(self, channel, data, *args, **kwargs):
        # for standard indicators, the default handle method will work
        # it creates instances of the storage class using the compute_value logic above
        super().handle(channel, data, *args, **kwargs)



##### MINIMALIST INDICATOR EXAMPLE #####

class FooStorage(IndicatorStorage):
    class_periods_list = [10, 20]
    requisite_pv_indexes = ["close_price", "close_volume"]

    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        foo_value = talib.FOO(requisite_pv_index_arrrays["close_price"], requisite_pv_index_arrrays["close_volume"], timeperiod=periods or self.periods)[-1]
        return str(foo_value) if not math.isnan(foo_value) else ""

    def produce_signal(self):
        self.send_signal(trend=(BULLISH if self.value > 1 else BEARISH))


class FooSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [PriceStorage, ]
    storage_class = ExampleIndicatorStorage

##### MINIMALIST INDICATOR EXAMPLE #####



from ticker import TickerStorage
from price import PriceStorage
from volume import VolumeStorage
from pv_history import PriceVolumeHistoryStorage
from ohlcv import OLHCVModel
from asset import AssetModel
from indicator import IndicatorStorage
from portfolio import PortfolioStorage



from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS, PERIODS_24HR
from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class HtTrendlineStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class HtTrendlineSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if self.index != 'close_price':
            logger.debug(f'index {self.index} is not close_price ...ignoring...')
            return

        new_ht_trendline_storage = HtTrendlineStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        results_dict = PriceStorage.query(
            ticker=self.ticker,
            exchange=self.exchange,
            index=self.index,
            periods_range=PERIODS_24HR*200
        )

        value_np_array = new_ht_trendline_storage.get_values_array_from_query(results_dict, limit=PERIODS_24HR*200)

        ht_trendline_value = talib.HT_TRENDLINE(value_np_array)[-1]
        # logger.debug(f'savingHt_trendline value {ht_trendline_value} for {self.ticker}')

        new_ht_trendline_storage.periods = PERIODS_24HR*200
        new_ht_trendline_storage.value = float(ht_trendline_value)
        new_ht_trendline_storage.save()



from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS
from apps.TA.storages.abstract.indicator import IndicatorStorage
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class ApoStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class ApoSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if str(self.index) is not "close_price":
            logger.debug(f'index {self.index} is not close_price ...ignoring...')
            return

        new_apo_storage = ApoStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        for horizon in HORIZONS:
            periods = horizon * 50

            close_value_np_array = new_apo_storage.get_denoted_price_array("close_price", periods)

            try:
                apo_value = talib.APO(close_value_np_array, fastperiod=12, slowperiod=26, matype=0)[-1]

                # logger.debug(f'savingApo value {apo_value} for {self.ticker} on {periods} periods')
                new_apo_storage.periods = periods
                new_apo_storage.value = float(apo_value)
                new_apo_storage.save()

            except Exception as e:
                logger.error(str(e))



import logging

from .db_key import DB_key
from ..redis_db import POPOTO_REDIS_DB

logger = logging.getLogger('POPOTO.Query')


class QueryException(Exception):
    pass


class Query:
    """
    an interface for db query operations using Popoto Models
    """
    model_class: 'Model'
    options: 'ModelOptions'

    def __init__(self, model_class: 'Model'):
        self.model_class = model_class
        self.options = model_class._meta

    def get(self, db_key: DB_key = None, redis_key: str = None, **kwargs) -> 'Model':
        if not db_key and not redis_key and all([key in kwargs for key in self.options.key_field_names]):
            db_key = self.model_class(**kwargs).db_key

        if db_key and not redis_key:
            redis_key = db_key.redis_key

        if redis_key:
            from ..models.encoding import decode_popoto_model_hashmap
            instance = decode_popoto_model_hashmap(self.model_class, POPOTO_REDIS_DB.hgetall(redis_key))

        else:
            instances = self.filter(**kwargs)
            if len(instances) > 1:
                raise QueryException(
                    f"{self.model_class.__name__} found more than one unique instance. Use `query.filter()`"
                )
            instance = instances[0] if len(instances) == 1 else None

        # or not hasattr(instance, 'db_key')
        return instance or None

    def keys(self, catchall=False, **kwargs) -> list:
        if catchall:
            logger.warning("{catchall} is for debugging purposes only. Not for use in production environment")
            return list(POPOTO_REDIS_DB.keys(f"*{self.model_class.__name__}*"))
        else:
            return list(POPOTO_REDIS_DB.smembers(self.model_class._meta.db_class_set_key.redis_key))

    def all(self) -> list:
        redis_db_keys_list = self.keys()
        return Query.get_many_objects(self.model_class, set(redis_db_keys_list))

    def filter_for_keys_set(self, **kwargs) -> list:
        db_keys_sets = []
        employed_kwargs_set = set()

        # todo: use redis.SINTER for keyfield exact match filters

        for field_name, field in self.options.fields.items():
            # intersection of field params and filter kwargs
            params_for_field = set(kwargs.keys()) & set(field.get_filter_query_params(field_name))
            if not params_for_field:
                continue
            logger.debug(f"query on {field_name} with {params_for_field}")

            logger.debug({k: kwargs[k] for k in params_for_field})
            key_set = field.__class__.filter_query(
                self.model_class, field_name, **{k: kwargs[k] for k in params_for_field}
            )
            db_keys_sets.append(key_set)
            employed_kwargs_set = employed_kwargs_set | params_for_field

        # raise error on additional unknown query parameters
        for unknown_query_param in set(kwargs) - employed_kwargs_set:
            if not any([
                unknown_query_param.startswith(field_name) for field_name in self.options.relationship_field_names
            ]):
                raise QueryException(f"Invalid filter parameters: {set(kwargs) - employed_kwargs_set}")

        logger.debug(db_keys_sets)
        if not len(db_keys_sets):
            return []
        return list(set.intersection(*db_keys_sets))

    def filter(self, **kwargs) -> list:
        """
           Access any and all filters for the fields on the model_class
           Run query using the given paramters
           return a list of model_class objects
        """
        limit = int(kwargs.pop('limit')) if 'limit' in kwargs else 0
        db_keys_set = self.filter_for_keys_set(**kwargs)
        if not len(db_keys_set):
            return []
        db_keys_set = db_keys_set[:limit] if limit > 0 else db_keys_set
        return Query.get_many_objects(self.model_class, db_keys_set)

    def count(self, **kwargs) -> int:
        if not len(kwargs):
            return int(POPOTO_REDIS_DB.scard(self.model_class._meta.db_class_set_key.redis_key) or 0)
        return len(self.filter_for_keys_set(**kwargs))  # maybe possible to refactor to use redis.SINTERCARD

    @classmethod
    def get_many_objects(cls, model: 'Model', db_keys: set) -> list:
        from .encoding import decode_popoto_model_hashmap
        pipeline = POPOTO_REDIS_DB.pipeline()
        for db_key in db_keys:
            pipeline.hgetall(db_key)
        hashes_list = pipeline.execute()
        if {} in hashes_list:
            logger.error("one or more redis keys points to missing objects")
        return [decode_popoto_model_hashmap(model, redis_hash) for redis_hash in hashes_list if redis_hash]



import logging
from src.popoto.models import Model
from ..redis_db import POPOTO_REDIS_DB
from src.popoto.exceptions import ModelException

logger = logging.getLogger(__name__)


class KeyValueException(ModelException):
    pass


class KeyValueModel(Model):
    """
    stores things in redis database given a key and value
    by default uses the instance class name as the key
    recommend to uniquely identify the instance with a key prefix or suffix
    prefixes are for more specific categories of objects (eg. mammal:human:woman:Lisa )
    suffixes are for specific attributes (eg. Lisa:eye_color, Lisa:age, etc)
    """

    _key_prefix: str = ""
    _key: str = ""
    _key_suffix: str = ""

    def __init__(self, *args, **kwargs):
        # build key
        self._key_prefix = kwargs.get('key_prefix', "")
        self._key = kwargs.get('key', self.__class__.__name__)
        self._key_suffix = kwargs.get('key_suffix', "")
        self._db_key = self.get_db_key(refresh=True)
        kwargs.pop('db_key', '')
        super().__init__(db_key=self._db_key, **kwargs)

    def __str__(self):
        return str(self.get_db_key())

    def get_db_key(self, refresh=False):
        if refresh or not self._db_key:
            self._db_key = self.compile_db_key(
                key_prefix=self._key_prefix,
                key=self._key,
                key_suffix=self._key_suffix
            )
            # todo: add this line for using env in key
            # + "" if SIMULATED_ENV == "PRODUCTION" else str(SIMULATED_ENV)
        return self._db_key

    @classmethod
    def compile_db_key(cls, key: str, key_prefix: str, key_suffix: str) -> str:
        key = key or cls.__name__
        # logging.debug(f"{key}, {key_prefix}, {key_suffix}")
        return str(
            f'{key_prefix.strip(":")}:' +
            f'{key.strip(":")}' +
            f':{key_suffix.strip(":")}'
        ).replace("::", ":").strip(":")

    def save(self, *args, **kwargs):
        self._db_key = self.get_db_key()
        super().save(*args, **kwargs)

    @classmethod
    def get(cls, db_key: str, instance_key: str = None):
        key_prefix, key_suffix = db_key.split(instance_key or cls.__name__)
        key_prefix = key_prefix.strip(":")
        key_suffix = key_suffix.strip(":")
        return cls(key=instance_key or cls.__name__, key_prefix=key_prefix, key_suffix=key_suffix)

    @property
    def value(self):
        self._db_key = self.get_db_key()
        return super().value

    def delete(self, *args, **kwargs):
        self._db_key = self.get_db_key()
        return super().delete(*args, **kwargs)

    def revert(self):
        self._db_key = self.get_db_key(refresh=True)
        return super().revert()



from settings import LOAD_TALIB
if LOAD_TALIB:
    import math, talib

from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH, BEARISH, OTHER
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class BbandsStorage(IndicatorStorage):

    class_periods_list = [5]
    requisite_pv_indexes = ["close_price"]

    def get_width(self):
        self.value = self.get_value()
        if self.value:
            [self.upperband_val, self.middleband_val, self.lowerband_val] = [float(val) for val in
                                                                             self.value.split(":")]
            self.width = (self.upperband_val - self.lowerband_val) / self.middleband_val
        else:
            self.width = None
        return self.width


    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        """
        with cls.requisite_pv_indexes set

        :param index_value_arrrays: a dict with keys matching requisite+pv_indexes and values from self.get_denoted_price_array()
        :param periods: number of periods to compute value for
        :return:
        """
        periods = periods or self.periods

        upperband, middleband, lowerband = talib.BBANDS(
            requisite_pv_index_arrrays["close_price"],
            timeperiod=periods,
            nbdevup=2, nbdevdn=2, matype=0
        )

        logger.debug(f"Bbands computed: {upperband[-1]}:{middleband[-1]}:{lowerband[-1]}")

        if math.isnan(sum([upperband[-1], middleband[-1], lowerband[-1]])): return ""

        return f"{upperband[-1]}:{middleband[-1]}:{lowerband[-1]}"


    def produce_signal(self):

        self.value = self.get_value()
        # Bbands value like f"{upperband_val}:{middleband_val}:{lowerband_val}"

        [self.upperband_val, self.middleband_val, self.lowerband_val] = [float(val) for val in self.value.split(":")]
        self.width = (self.upperband_val - self.lowerband_val) / self.middleband_val

        query_result = BbandsStorage.query(
            ticker=self.ticker, exchange=self.exchange, timestamp=self.unix_timestamp, periods=180
        )
        if query_result['values_count'] < 180:
            return  # not enough data

        bbands_values = query_result['values'][-180:]

        upper, middle, lower = 0, 1, 2

        bbands_widths = [(float(v.split(":")[upper]) - float(v.split(":")[lower])) / float(v.split(":")[middle])
                         for v in query_result['values']]

        if self.width <= min(bbands_widths):  # smallest width (squeeze) in the last 180 periods
            # squeeze = True

            self.price = float(PriceStorage.query(ticker=self.ticker, exchange=self.exchange, index="close_price",
                                                  timestamp=self.unix_timestamp)['values'][-1])

            if self.price > self.upperband_val:  # price breaks out above the band
                self.trend = BULLISH

            elif self.price < self.lowerband_val:  # price breaks out below the band
                self.trend = BEARISH

            else:  # price doesn't break out - but the squeeze thing is cool
                self.trend = OTHER

            self.send_signal(type="BBands", trend=self.trend, width=self.width)
            logger.debug("new BBands signal sent")


class BbandsSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [PriceStorage]
    storage_class = BbandsStorage



import logging

from src.popoto.exceptions import FinanceException
from ...finance import TickerStorage, PriceVolumeHistoryStorage, TickerSubscriber
# score_is_near_5min, clear_pv_history_values, PRICE_INDEXES, generate_pv_storages


logger = logging.getLogger(__name__)


class PriceException(FinanceException):
    pass


class PriceStorage(TickerStorage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = kwargs.get('index', "close_price")
        self.value = kwargs.get('value')
        self.db_key_suffix = f':{self.index}'

    def save(self, *args, **kwargs):

        # meets basic requirements for saving
        if not all([self.ticker, self.publisher,
                    self.index, self.value,
                    self.unix_timestamp]):
            logger.error("incomplete information, cannot save \n" + str(self.__dict__))
            raise PriceException("save error, missing data")

        if not self.force_save:
            if self.index not in PRICE_INDEXES:
                logger.error("price index not in approved list, raising exception...")
                raise PriceException("unknown index")

        if self.unix_timestamp % 3600 != 0:
            raise PriceException("price timestamp should be % 3600")

        self.db_key_suffix = self.db_key_suffix or f':{self.index}'
        return super().save(*args, **kwargs)

    @classmethod
    def query(cls, *args, **kwargs):

        if kwargs.get("periods_key", None):
            raise PriceException("periods_key is not usable in PriceStorage query")

        key_suffix = kwargs.get("key_suffix", "")
        index = kwargs.get("index", "close_price")
        kwargs["key_suffix"] = f'{index}' + (f':{key_suffix}' if key_suffix else "")

        results_dict = super().query(*args, **kwargs)

        results_dict['index'] = index
        return results_dict


class PriceSubscriber(TickerSubscriber):
    classes_subscribing_to = [
        PriceVolumeHistoryStorage
    ]

    def handle(self, channel, data, *args, **kwargs):
        # parse data like...
        # {
        #     "key": "POLY_BTC:binance:PriceVolumeHistoryStorage:open_price",
        #     "name": "3665:176255.873",
        #     "score": "176255.873"
        # }

        # eg. sorted_set_key = data["key"]

        [ticker, publisher, object_class, index] = data["key"].split(":")
        if not object_class == channel == PriceVolumeHistoryStorage.__name__:
            logger.warning(f'Unexpected that these are not the same:'
                           f'object_class: {object_class}, '
                           f'channel: {channel}, '
                           f'subscribing class: {PriceVolumeHistoryStorage.__name__}')
        [value, name_score] = data["name"].split(":")

        score = float(data["score"])

        if not float(name_score) == float(data["score"]):
            logger.warning(f'Unexpected that score in name {name_score}'
                           f'is different than score {score}')

        if score_is_near_5min(score):
            if generate_pv_storages(ticker, publisher, index, score):
                if index == "close_price":
                    clear_pv_history_values(ticker, publisher, score)



import logging
from datetime import datetime
import numpy as np
from src.popoto._archive.key_value import KeyValueModel
from src.popoto.redis_db import POPOTO_REDIS_DB, BEGINNING_OF_TIME
from src.popoto.exceptions import ModelException
logger = logging.getLogger(__name__)


class TimeseriesException(ModelException):
    pass


class TimeseriesModel(KeyValueModel):
    """
    stores things in a sorted set unique to each ticker and publisher
    todo: split the db by each publisher source

    """
    class_describer = "timeseries"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # 'timestamp' REQUIRED, VALIDATE
        try:
            self.unix_timestamp = int(kwargs['timestamp'])  # int eg. 1483228800
        except KeyError:
            raise TimeseriesException("timestamp required for TimeseriesModel objects")
        except ValueError:
            raise TimeseriesException(
                "timestamp must be castable as integer, received {ts}".format(
                    ts=kwargs.get('timestamp')))
        except Exception as e:
            raise TimeseriesException(str(e))

        if self.unix_timestamp < BEGINNING_OF_TIME:
            raise TimeseriesException(f"timestamp {self.unix_timestamp} is before begining of time {BEGINNING_OF_TIME}")

    def save_own_existance(self, describer_key=""):
        self.describer_key = describer_key or f'{self.__class__.class_describer}:{self._db_key}'

    @classmethod
    def score_from_timestamp(cls, timestamp) -> float:
        return round((float(timestamp) - BEGINNING_OF_TIME) / 300, 3)

    @classmethod
    def timestamp_from_score(cls, score) -> int:
        return int(float(score) * 300) + BEGINNING_OF_TIME

    @classmethod
    def datetime_from_score(cls, score) -> datetime:
        return datetime.fromtimestamp(cls.timestamp_from_score(score))

    @classmethod
    def periods_from_seconds(cls, seconds) -> float:
        return float(seconds) / 300

    @classmethod
    def seconds_from_periods(cls, periods) -> int:
        return int(float(periods) * 300)

    @classmethod
    def query(cls, key: str = "", key_suffix: str = "", key_prefix: str = "",
              timestamp: int = None,
              timestamp_tolerance: int = 299,
              periods_range: float = 0.01,
              *args, **kwargs) -> dict:
        """
        :param key: the exact redis sortedset key (optional)
        :param key_suffix: suffix on the key  (optional)
        :param key_prefix: prefix on the key (optional)
        :param timestamp: timestamp for most recent value returned  (optional, default returns latest)
        :param periods_range: number of periods desired in results (optional, default 0, so only return 1 value)
        :param timestamp_tolerance: tolerance in seconds on results within timestamp and period range (optional, defualt=299)
        :return: dict(values=[], ...)
        """

        sorted_set_key = cls.compile_db_key(key=key, key_prefix=key_prefix, key_suffix=key_suffix)
        # logger.debug(f'query for sorted set key {sorted_set_key}')
        # example key f'{key_prefix}:{cls.__name__}:{key_suffix}'

        # do a quick check to make sure this is a class of things we know is in existence
        describer_key = f'{cls.class_describer}:{sorted_set_key}'
        # if no timestamp, assume query to find the most recent, the last one

        if not timestamp:
            query_response = POPOTO_REDIS_DB.zrange(sorted_set_key, -1, -1)
            try:
                [value, timestamp] = query_response[0].decode("utf-8").split(":")
            except:
                value, timestamp = "unknown", BEGINNING_OF_TIME

            min_score = max_score = cls.score_from_timestamp(timestamp)

        else:
            # compress timestamps to scores
            target_score = cls.score_from_timestamp(timestamp)
            score_tolerance = cls.periods_from_seconds(timestamp_tolerance)

            # logger.debug(f"querying for key {sorted_set_key} with score {target_score} and back {periods_range} periods")

            min_score, max_score = (target_score - score_tolerance - periods_range), (target_score + score_tolerance)

            query_response = POPOTO_REDIS_DB.zrangebyscore(sorted_set_key, min_score, max_score)

        # OLD example query_response = [b'0.06288:1532163247']
        # which came from f'{self.value}:{str(self.unix_timestamp)}'

        # NEW example query_response = [b'0.06288:1532163247']
        # which came from f'{self.value}:{str(score)}' where score = (self.timestamp-BEGINNING_OF_TIME)/300

        return_dict = {
            'values': [],
            'values_count': 0,
            'timestamp': timestamp,
            'earliest_timestamp': cls.timestamp_from_score(min_score),
            'latest_timestamp': cls.timestamp_from_score(max_score),
            'periods_range': periods_range,
            'period_size': 300,
        }

        if not len(query_response):
            return return_dict

        try:
            return_dict['values_count'] = len(query_response)

            if len(query_response) < periods_range + 1:
                return_dict["warning"] = "fewer values than query's periods_range"

            values = [value_score.decode("utf-8").split(":")[0] for value_score in query_response]
            scores = [float(value_score.decode("utf-8").split(":")[1]) for value_score in query_response]
            # todo: double check that [-1] in list is most recent timestamp

            return_dict.update({
                'values': values,
                'scores': scores,
                'earliest_timestamp': cls.timestamp_from_score(float(scores[0])),
                'latest_timestamp': cls.timestamp_from_score(float(scores[-1])),
            })
            return return_dict

        except IndexError:
            return return_dict

        except Exception as e:
            logger.error("redis query problem: " + str(e))
            return {'error': "redis query problem: " + str(e),  # wtf happened?
                    'values': []}

    @staticmethod
    def get_values_array_from_query(query_results: dict, limit: int = 0):

        value_array = [float(v) for v in query_results['values']]

        if limit:
            if not isinstance(limit, int) or limit < 1:
                raise TimeseriesException(f"bad limit: {limit}")

            elif len(value_array) > limit:
                value_array = value_array[-limit:]

        return np.array(value_array)

    def get_z_add_data(self):
        z_add_key = f'{self.get_db_key()}'  # set key name
        z_add_score = f'{self.score_from_timestamp(self.unix_timestamp)}'  # timestamp as score (int or float)
        z_add_name = f'{self.value}:{z_add_score}'  # item unique value
        z_add_data = {"key": z_add_key, "name": z_add_name, "score": z_add_score}  # key, score, name
        return z_add_data

    def save(self, publish=False, pipeline=None, *args, **kwargs):
        if self.value is None:
            raise ModelException("no value set, nothing to save!")
        if not self.force_save:
            # validate some rules here?
            pass

        self.save_own_existance()

        z_add_data = self.get_z_add_data()
        # logger.debug(f'saving data with args {z_add_data}')

        if pipeline is not None:
            pipeline = pipeline.zadd(z_add_data['key'], {z_add_data['name']: z_add_data['score']})
            # logger.debug("added command to redis pipeline")
            if publish:
                pipeline = self.publish(pipeline)
            return pipeline
        else:
            response = POPOTO_REDIS_DB.zadd(z_add_data['key'], {z_add_data['name']: z_add_data['score']})
            # logger.debug("no pipeline, executing zadd command immediately.")
            if publish:
                self.publish()
            return response

    def publish(self, pipeline=None):
        return super().publish(data=self.get_z_add_data(), pipeline=pipeline)

    def get_value(self, *args, **kwargs):
        TimeseriesException("function not yet implemented! ¯\_(ツ)_/¯ ")
        pass


"""
We can scan the newest or oldest event ids with ZRANGE 4,
maybe later pulling the events themselves for analysis.

We can get the 10 or even 100 events immediately
before or after a timestamp with ZRANGEBYSCORE
combined with the LIMIT argument.

We can count the number of events that occurred
in a specific time period with ZCOUNT.

https://www.infoq.com/articles/redis-time-series
"""



import redis
from .field import Field
import logging

from ..models.db_key import DB_key
from ..models.query import QueryException
from ..redis_db import POPOTO_REDIS_DB

logger = logging.getLogger('POPOTO.Relationship')


class Relationship(Field):
    """
    A field that stores references to one or more other model instances.
    """
    type: 'Model' = None
    model: 'Model' = None
    many: bool = False
    null: bool = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from ..models.base import Model
        relationship_field_defaults = {
            'type': Model,
            'model': None,
            'many': False,
            'null': True,
        }
        self.field_defaults.update(relationship_field_defaults)
        # set field options, let kwargs override
        for k, v in relationship_field_defaults.items():
            setattr(self, k, kwargs.get(k, v))


    def get_filter_query_params(self, field_name):
        return super().get_filter_query_params(field_name) + [
            f'{field_name}',
        ] + [
            f'{field_name}__{field.get_filter_query_params(related_field_name)}'
            for related_field_name, field in self.model._meta.fields.items()
            if not isinstance(field, Relationship)  # not ready for recursive compilation
        ]

    @classmethod
    def on_save(cls, model_instance: 'Model', field_name: str, field_value: 'Model', pipeline=None, **kwargs):
        from ..models.base import Model
        # on a one-to-many, save the set of many with the related instance
        # add this instance's id to a relationship set based on the related model
        # example: "$RelationshipF:Membership:person:person_db_key"
        relationship_set_db_key = DB_key(
            cls.get_special_use_field_db_key(model_instance, field_name),
            field_value.db_key if field_value else "None"
        )

        if pipeline and field_value is None:
            return pipeline.srem(relationship_set_db_key.redis_key, model_instance.db_key.redis_key)
        elif pipeline and isinstance(field_value, Model):
            return pipeline.sadd(relationship_set_db_key.redis_key, model_instance.db_key.redis_key)
        elif field_value is None:
            return POPOTO_REDIS_DB.srem(relationship_set_db_key.redis_key, model_instance.db_key.redis_key)
        elif isinstance(field_value, Model):
            return POPOTO_REDIS_DB.sadd(relationship_set_db_key.redis_key, model_instance.db_key.redis_key)
        else:
            return pipeline if pipeline else None

    @classmethod
    def on_delete(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None, **kwargs):
        # todo: it's possible this instance is not fully loaded or has been changed.
        #  Need to reload from db before deleting
        #  Example: on person.friend.delete() the person.friend.friend will have field_value as a keystring
        #  It will not be a model instance, so this method will fail on field_value.db_key below
        relationship_set_db_key = DB_key(
            cls.get_special_use_field_db_key(model_instance, field_name),
            field_value.db_key if field_value else "None"
        )
        if pipeline:
            return pipeline.srem(relationship_set_db_key.redis_key, model_instance.db_key.redis_key)
        else:
            return POPOTO_REDIS_DB.srem(relationship_set_db_key.redis_key, model_instance.db_key.redis_key)

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """
        from ..models.base import Model

        keys_lists_to_intersect = list()
        pipeline = POPOTO_REDIS_DB.pipeline()

        for query_param, query_value in query_params.items():

            if query_param == f'{field_name}':
                if not isinstance(query_value, Model):
                    raise QueryException(f"Query filter on Relationship expects model instance. Instead, got {query_value}")

                relationship_set_db_key = DB_key(cls.get_special_use_field_db_key(model, field_name), query_value.db_key)
                keys_lists_to_intersect.append(
                    POPOTO_REDIS_DB.smembers(relationship_set_db_key.redis_key)
                )

            elif query_param.startswith(f'{field_name}__'):
                field = model._meta.fields[field_name]
                relationship_field_name = query_param.strip(f'{field_name}__').split("__")[0]

                relationship_field_values = field.model._meta.fields[relationship_field_name].filter_query(
                    model=field.model, field_name=relationship_field_name,
                    **{query_param.strip(f'{field_name}__'): query_value}
                )
                # note: this will be recursive if references another relationship and so forth

                for relationship_field_value in relationship_field_values:
                    relationship_set_db_key = DB_key(
                        cls.get_special_use_field_db_key(model, field_name),
                        relationship_field_value.db_key
                    )
                    pipeline.smembers(relationship_set_db_key.redis_key)

        keys_lists_to_intersect += pipeline.execute()
        logger.debug(keys_lists_to_intersect)
        if len(keys_lists_to_intersect):
            return set.intersection(*[set(key_list) for key_list in keys_lists_to_intersect])
        return set()




from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA.storages.abstract.indicator import IndicatorStorage
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class BopStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class BopSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if self.index is not 'low_price':
            logger.debug(f'index {self.index} is not `low_price` ...ignoring...')
            return

        new_bop_storage = BopStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)


        periods = 2  # doesn't matter, just enough to grab the last one

        open_value_np_array = new_bop_storage.get_denoted_price_array("open_price", periods)
        high_value_np_array = new_bop_storage.get_denoted_price_array("high_price", periods)
        low_value_np_array = new_bop_storage.get_denoted_price_array("low_price", periods)
        close_value_np_array = new_bop_storage.get_denoted_price_array("close_price", periods)

        bop_value = talib.BOP(open_value_np_array, high_value_np_array, low_value_np_array, close_value_np_array)[-1]
        # logger.debug(f'savingBop value {bop_value} for {self.ticker} on {periods} periods')

        new_bop_storage.value = bop_value
        new_bop_storage.save()




def round_sig_figs(value, sig_figs: int) -> float:
    value = float(value)
    non_decimal_place = len(str(int(value)))
    # decimal_places = max(len(str(value % 1))-2, 0)

    if value < 0:
        distance_from_decimal = 0
        for char in str(value):
            if char is ".": continue
            elif int(char) > 0: break
            else: distance_from_decimal += 1
        return round(value, (distance_from_decimal+sig_figs))
    else:
        return round(value, (sig_figs-non_decimal_place))



from settings import LOAD_TALIB
if LOAD_TALIB:
    import math, talib

from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH, BEARISH, OTHER
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger

SMA_LIST = [9, 20, 26, 30, 50, 52, 60, 120, 200]


class SmaStorage(IndicatorStorage):
    # example sorted_set_key = "BTC_USDT:poloniex:SmaStorage:20"

    class_periods_list = SMA_LIST
    requisite_pv_indexes = ["close_price"]

    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        """
        with cls.requisite_pv_indexes set

        :param index_value_arrrays: a dict with keys matching requisite+pv_indexes and values from self.get_denoted_price_array()
        :param periods: number of periods to compute value for
        :return:
        """
        periods = periods or self.periods
        sma_value = talib.SMA(requisite_pv_index_arrrays["close_price"], timeperiod=periods)[-1]

        logger.debug(f"SMA computed: {sma_value}")

        if math.isnan(sma_value):
            return ""

        return str(sma_value)


    def produce_signal(self):
        """
        defining the criteria for sending signals

        :return: None
        """
        if "this indicator" == "interesting":
            self.send_signal(trend=BULLISH)


class SmaSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [PriceStorage]
    storage_class = SmaStorage



import logging

from src.popoto._archive.timeseries import TimeseriesException, TimeseriesModel

logger = logging.getLogger(__name__)


class PortfolioException(TimeseriesException):
    pass


class PortfolioStorage(TimeseriesModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = str(kwargs.get('id'))
        self.value = kwargs.get('value')
        self.db_key_suffix = f':{self.id}'

    def save(self, *args, **kwargs):

        # meets basic requirements for saving
        if not all([self.id, self.value is not None, self.unix_timestamp]):
            logger.error("incomplete information, cannot save \n" + str(self.__dict__))
            raise PortfolioException("save error, missing data")

        if self.unix_timestamp % 3600 != 0:  # on the hour
            raise PortfolioException("price timestamp should be % 3600")

        return super().save(*args, **kwargs)

    @classmethod
    def query(cls, *args, **kwargs):

        key_suffix = kwargs.get("key_suffix", "")
        id = kwargs.get("id", "close_price")
        kwargs["key_suffix"] = f'{id}' + (f':{key_suffix}' if key_suffix else "")

        results_dict = super().query(*args, **kwargs)
        results_dict['id'] = id
        return results_dict



import logging

import redis

from .encoding import encode_popoto_model_obj
from .db_key import DB_key
from .query import Query
from ..fields.field import Field, VALID_FIELD_TYPES
from ..fields.key_field_mixin import KeyFieldMixin
from ..fields.sorted_field_mixin import SortedFieldMixin
from ..fields.geo_field import GeoField
from ..fields.relationship import Relationship
from ..redis_db import POPOTO_REDIS_DB

logger = logging.getLogger('POPOTO.model_base')

global RELATED_MODEL_LOAD_SEQUENCE
RELATED_MODEL_LOAD_SEQUENCE = set()


class ModelException(Exception):
    pass


class ModelOptions:
    def __init__(self, model_name):
        self.model_name = model_name
        self.hidden_fields = dict()
        self.explicit_fields = dict()
        self.key_field_names = set()
        # self.auto_field_names = set()
        # self.list_field_names = set()
        # self.set_field_names = set()
        self.relationship_field_names = set()
        self.sorted_field_names = set()
        self.geo_field_names = set()

        # todo: should this be a dict of related objects or just a list of field names?
        # self.related_fields = {}  # model becomes graph node

        # todo: allow customizing this in model.Meta class
        self.db_class_key = DB_key(self.model_name)
        self.db_class_set_key = DB_key("$Class", self.db_class_key)

        self.abstract = False
        self.unique_together = []
        self.index_together = []
        self.parents = []
        self.auto_created = False
        self.base_meta = None

    def add_field(self, field_name: str, field: Field):
        if field_name.startswith("_") and field_name not in self.hidden_fields:
            self.hidden_fields[field_name] = field
        elif field_name not in self.explicit_fields:
            self.explicit_fields[field_name] = field
        else:
            raise ModelException(f"{field_name} is already a Field on the model")

        if isinstance(field, KeyFieldMixin):
            self.key_field_names.add(field_name)
        # if field.auto:
        #     self.auto_field_names.add(field_name)
        if isinstance(field, SortedFieldMixin):
            self.sorted_field_names.add(field_name)
        if isinstance(field, GeoField):
            self.geo_field_names.add(field_name)
        # elif isinstance(field, ListField):
        #     self.list_field_names.add(field_name)
        if isinstance(field, Relationship):
            self.relationship_field_names.add(field_name)

    @property
    def fields(self) -> dict:
        return {**self.explicit_fields, **self.hidden_fields}

    @property
    def field_names(self) -> list:
        return list(self.fields.keys())

    @property
    def db_key_length(self):
        return 1 + len(self.key_field_names)

    def get_db_key_index_position(self, field_name):
        return 1 + sorted(self.key_field_names).index(field_name)


class ModelBase(type):
    """Metaclass for all Popoto Models."""

    def __new__(cls, name, bases, attrs, **kwargs):

        # Initialization is only performed for a Model and its subclasses
        parents = [b for b in bases if isinstance(b, ModelBase)]
        if not parents:
            return super().__new__(cls, name, bases, attrs, **kwargs)

        # logger.debug({k: v for k, v in attrs.items() if not k.startswith('__')})
        module = attrs.pop('__module__')
        new_attrs = {'__module__': module}
        attr_meta = attrs.pop('Meta', None)
        options = ModelOptions(name)
        options.parents = parents

        for obj_name, obj in attrs.items():
            if obj_name.startswith("__"):
                # builtin or inherited private vars and methods
                new_attrs[obj_name] = obj

            elif isinstance(obj, Field):
                # save field instance
                # attr will be overwritten as a field.type
                # model will handle this and set default values
                options.add_field(obj_name, obj)

            elif callable(obj) or hasattr(obj, '__func__') or hasattr(obj, '__set__'):
                # a callable method or property
                new_attrs[obj_name] = obj

            elif obj_name.startswith("_"):
                # a private static attr not to be saved in the db
                new_attrs[obj_name] = obj

            else:
                raise ModelException(
                    f"public model attributes must inherit from class Field. "
                    f"Try using a private var (eg. _{obj_name})_"
                )

        # todo: handle multiple inheritance
        # for base in parents:
        #     for field_name, field in base.auto_fields.items():
        #         options.add_field(field_name, field)

        new_class = super().__new__(cls, name, bases, new_attrs)

        options.abstract = getattr(attr_meta, 'abstract', False)
        options.meta = attr_meta or getattr(new_class, 'Meta', None)
        options.base_meta = getattr(new_class, '_meta', None)
        new_class._meta = options
        new_class.objects = new_class.query = Query(new_class)
        return new_class


class Model(metaclass=ModelBase):
    query: Query

    def __init__(self, **kwargs):
        cls = self.__class__
        # self._ttl = kwargs.get('ttl', None)
        # self._expire_at = kwargs.get('expire_at', None)

        # allow init kwargs to set any base parameters
        self.__dict__.update(kwargs)

        # add auto KeyField if needed
        if not len(self._meta.key_field_names):
            from ..fields.shortcuts import AutoKeyField
            self._meta.add_field('_auto_key', AutoKeyField())

        # prep AutoKeys with new default ids
        for field in self._meta.fields.values():
            if hasattr(field, 'auto') and field.auto:
                field.set_auto_key_value()

        # set defaults
        for field_name, field in self._meta.fields.items():
            setattr(self, field_name, field.default)

        # set field values from init kwargs
        for field_name in self._meta.fields.keys() & kwargs.keys():
            setattr(self, field_name, kwargs.get(field_name))

        # load relationships
        if len(self._meta.relationship_field_names):
            global RELATED_MODEL_LOAD_SEQUENCE
            is_parent_model = len(RELATED_MODEL_LOAD_SEQUENCE) == 0
            for field_name in self._meta.relationship_field_names:
                if f"{self.__class__.__name__}.{field_name}" in RELATED_MODEL_LOAD_SEQUENCE:
                    continue
                RELATED_MODEL_LOAD_SEQUENCE.add(f"{self.__class__.__name__}.{field_name}")

                field_value = getattr(self, field_name)
                if isinstance(field_value, Model):
                    setattr(self, field_name, field_value)
                elif isinstance(field_value, str):
                    setattr(
                        self, field_name,
                        self._meta.fields[field_name].model.query.get(redis_key=field_value)
                    )

                # todo: lazy load the instance from the db
                elif not field_value:
                    setattr(self, field_name, None)
                else:
                    raise ModelException(f"{field_name} expects model instance or redis_key")
            if is_parent_model:
                RELATED_MODEL_LOAD_SEQUENCE = set()

        self._ttl = None  # todo: set default in child Meta class
        self._expire_at = None  # todo: datetime? or timestamp?

        # validate initial attributes
        if not self.is_valid(null_check=False):  # exclude null, will validate null values on pre-save
            raise ModelException(f"Could not instantiate class {self}")

        self._redis_key = None
        # _db_key used by Redis cannot be known without performance cost
        # _db_key is predicted until synced during save() call
        if None not in [getattr(self, key_field_name) for key_field_name in self._meta.key_field_names]:
            self._redis_key = self.db_key.redis_key
        self.obsolete_redis_key = None  # to be used when db_key changes between loading and saving the object
        self._db_content = dict()  # empty until synced during save() call

        # todo: create set of possible custom field keys

    @property
    def db_key(self) -> DB_key:
        """
        the db key must include the class name - equivalent to db table name
        keys append alphabetically.
        if another order is required, propose feature request in GitHub issue
        possible solutions include param on each model's KeyField order=int
        OR model Meta: key_order = [keyname, keyname, ]
        OR both
        """
        return DB_key(self._meta.db_class_key, [
            getattr(self, key_field_name) or "None"
            for key_field_name in sorted(self._meta.key_field_names)
        ])

    def __repr__(self):
        return f"<{self.__class__.__name__} Popoto object at {self.db_key.redis_key}>"

    def __str__(self):
        return str(self.db_key)

    def __eq__(self, other):
        """
        equality method
        instances with the same key(s) and class are considered equal
        except when any key(s) are None, they are not equal to anything except themselves.

        for evaluating all instance values against each other, use something like this:
        self_dict = self._meta.fields.update((k, self.__dict__[k]) for k in set(self.__dict__).intersection(self._meta.fields))
        other_dict = other._meta.fields.update((k, other.__dict__[k]) for k in set(other.__dict__).intersection(other._meta.fields))
        return repr(dict(sorted(self_dict))) == repr(dict(sorted(other_dict)))
        """
        try:
            if isinstance(other, self.__class__):
                # always False if if any KeyFields are None
                if (None in [
                    self._meta.fields.get(key_field_name) for key_field_name in self._meta.key_field_names
                ]) or (None in [
                    other._meta.fields.get(key_field_name) for key_field_name in other._meta.key_field_names
                ]):
                    return repr(self) == repr(other)
                return self.db_key == other.db_key
        except:
            return False
        else:
            return False

    # @property
    # def field_names(self):
    #     return [
    #         k for k, v in self.__dict__.items()
    #         if all([not k.startswith("_"), k + "_meta" in self.__dict__])
    #     ]

    def is_valid(self, null_check=True) -> bool:
        """
            todo: validate values
            - field.type ✅
            - field.null ✅
            - field.max_length ✅
            - ttl, expire_at - todo
        """

        for field_name in self._meta.field_names:
            # type check the field values against their class specified type, unless null/None

            if all([
                getattr(self, field_name) is not None,
                not isinstance(getattr(self, field_name), self._meta.fields[field_name].type)
            ]):
                try:
                    if getattr(self, field_name) is not None:
                        if self._meta.fields[field_name].type in VALID_FIELD_TYPES:
                            setattr(self, field_name, self._meta.fields[field_name].type(getattr(self, field_name)))
                        else:
                            pass  # do not force typing if custom type is defined
                    if not isinstance(getattr(self, field_name), self._meta.fields[field_name].type):
                        raise TypeError(f"Expected {field_name} to be type {self._meta.fields[field_name].type}. "
                                        f"It is type {type(getattr(self, field_name))}")
                except TypeError as e:
                    logger.error(
                        f"{str(e)} \n Change the value or modify type on {self.__class__.__name__}.{field_name}"
                    )
                    return False

            # check non-nullable fields
            if null_check and \
                    self._meta.fields[field_name].null is False and \
                    getattr(self, field_name) is None:
                error = f"{field_name} is None/null. " \
                        f"Set a value or set null=True on {self.__class__.__name__}.{field_name}"
                logger.error(error)
                return False

            # validate str max_length
            if self._meta.fields[field_name].type == str and \
                    getattr(self, field_name) and \
                    len(getattr(self, field_name)) > self._meta.fields[field_name].max_length:
                error = f"{field_name} is greater than max_length={self._meta.fields[field_name].max_length}"
                logger.error(error)
                return False

            if self._ttl and self._expire_at:
                raise ModelException("Can set either ttl and expire_at. Not both.")

        for field_name, field_value in self.__dict__.items():
            if field_name in self._meta.fields.keys():
                field_class = self._meta.fields[field_name].__class__
                if not field_class.is_valid(self._meta.fields[field_name], field_value, null_check=null_check):
                    error = f"Validation on [{field_name}] Field failed"
                    logger.error(error)
                    return False

        return True

    def pre_save(self, pipeline: redis.client.Pipeline = None,
                 ignore_errors: bool = False, **kwargs):
        """
        Model instance preparation for saving.
        """
        if not self.is_valid():
            error_message = "Model instance parameters invalid. Failed to save."
            if ignore_errors:
                logger.error(error_message)
            else:
                raise ModelException(error_message)
            return False

        # run any necessary formatting on field data before saving
        for field_name, field in self._meta.fields.items():
            setattr(
                self, field_name,
                field.format_value_pre_save(getattr(self, field_name))
            )
        return pipeline if pipeline else True

    def save(self, pipeline: redis.client.Pipeline = None,
             ttl=None, expire_at=None, ignore_errors: bool = False,
             **kwargs):
        """
            Model instance save method. Uses Redis HSET command with key, dict of values, ttl.
            Also triggers all field on_save methods.
        """

        pipeline_or_success = self.pre_save(pipeline=pipeline, ignore_errors=ignore_errors, **kwargs)
        if not pipeline_or_success:
            return pipeline or False
        elif pipeline:
            pipeline = pipeline_or_success

        new_db_key = DB_key(self.db_key)  # todo: why have a new key??
        if self._redis_key != new_db_key.redis_key:
            self.obsolete_redis_key = self._redis_key

        # todo: implement and test tll, expire_at
        ttl, expire_at = (ttl or self._ttl), (expire_at or self._expire_at)

        """
        1. save object as hashmap
        2. optionally set ttl, expire_at
        3. add to class set
        4. if obsolete key, delete and run field on_delete methods
        5. run field on_save methods
        6. save private version of compiled db key
        """

        hset_mapping = encode_popoto_model_obj(self)  # 1
        self._db_content = hset_mapping  # 1

        if isinstance(pipeline, redis.client.Pipeline):
            pipeline = pipeline.hset(new_db_key.redis_key, mapping=hset_mapping)  # 1
            # if ttl is not None:
            #     pipeline = pipeline.expire(new_db_key, ttl)  # 2
            # if expire_at is not None:
            #     pipeline = pipeline.expire_at(new_db_key, expire_at)  # 2
            pipeline = pipeline.sadd(self._meta.db_class_set_key.redis_key, new_db_key.redis_key)  # 3
            if self.obsolete_redis_key and self.obsolete_redis_key != new_db_key.redis_key:  # 4
                for field_name, field in self._meta.fields.items():
                    pipeline = field.on_delete(  # 4
                        model_instance=self, field_name=field_name,
                        field_value=getattr(self, field_name),
                        pipeline=pipeline, **kwargs
                    )
                pipeline.delete(self.obsolete_redis_key)  # 4
                self.obsolete_redis_key = None
            for field_name, field in self._meta.fields.items():  # 5
                pipeline = field.on_save(  # 5
                    self, field_name=field_name,
                    field_value=getattr(self, field_name),
                    # ttl=ttl, expire_at=expire_at,
                    ignore_errors=ignore_errors,
                    pipeline=pipeline, **kwargs
                )
            self._redis_key = new_db_key.redis_key  # 6
            return pipeline

        else:
            db_response = POPOTO_REDIS_DB.hset(new_db_key.redis_key, mapping=hset_mapping)  # 1
            # if ttl is not None:
            #     POPOTO_REDIS_DB.expire(new_db_key, ttl)  # 2
            # if expire_at is not None:
            #     POPOTO_REDIS_DB.expireat(new_db_key, ttl)  # 2
            POPOTO_REDIS_DB.sadd(self._meta.db_class_set_key.redis_key, new_db_key.redis_key)  # 2

            if self.obsolete_redis_key and self.obsolete_redis_key != new_db_key.redis_key:  # 4
                for field_name, field in self._meta.fields.items():
                    field.on_delete(  # 4
                        model_instance=self, field_name=field_name,
                        field_value=getattr(self, field_name),
                        pipeline=None, **kwargs
                    )
                POPOTO_REDIS_DB.delete(self.obsolete_redis_key)  # 4
                self.obsolete_redis_key = None

            for field_name, field in self._meta.fields.items():  # 5
                field.on_save(  # 5
                    self, field_name=field_name,
                    field_value=getattr(self, field_name),
                    # ttl=ttl, expire_at=expire_at,
                    ignore_errors=ignore_errors,
                    pipeline=None, **kwargs
                )

            self._redis_key = new_db_key.redis_key  # 6
            return db_response

    @classmethod
    def create(cls, pipeline: redis.client.Pipeline = None, **kwargs):
        instance = cls(**kwargs)
        pipeline_or_db_response = instance.save(pipeline=pipeline)
        return pipeline_or_db_response if pipeline else instance

    @classmethod
    def load(cls, db_key: str = None, **kwargs):
        return cls.query.get(db_key=db_key or cls(**kwargs).db_key)

    def delete(self, pipeline: redis.client.Pipeline = None, *args, **kwargs):
        """
            Model instance delete method. Uses Redis DELETE command with key.
            Also triggers all field on_delete methods.
        """

        delete_redis_key = self._redis_key or self.db_key.redis_key

        """
        1. delete object as hashmap
        2. delete from class set
        3. run field on_delete methods
        4. reset private vars
        """

        if pipeline is not None:
            pipeline = pipeline.delete(delete_redis_key)  # 1
            pipeline = pipeline.srem(self._meta.db_class_set_key.redis_key, delete_redis_key)  # 2

            for field_name, field in self._meta.fields.items():  # 3
                pipeline = field.on_delete(  # 3
                    model_instance=self, field_name=field_name,
                    field_value=getattr(self, field_name),
                    pipeline=pipeline, **kwargs
                )

            self._db_content = dict()  # 4
            return pipeline

        else:
            db_response = POPOTO_REDIS_DB.delete(delete_redis_key)  # 1
            POPOTO_REDIS_DB.srem(self._meta.db_class_set_key.redis_key, delete_redis_key)  # 2

            for field_name, field in self._meta.fields.items():  # 3
                field.on_delete(  # 3
                    model_instance=self, field_name=field_name,
                    field_value=getattr(self, field_name),
                    pipeline=None, **kwargs
                )

            self._db_content = dict()  # 4
            return bool(db_response > 0)

    @classmethod
    def get_info(cls):
        from itertools import chain
        query_filters = list(chain(*[
            field.get_filter_query_params(field_name)
            for field_name, field in cls._meta.fields.items()
        ]))
        return {
            'name': cls.__name__,
            'fields': cls._meta.field_names,
            'query_filters': query_filters,
        }



from collections import namedtuple
import redis
from .field import Field
import logging
from ..models.db_key import DB_key

logger = logging.getLogger('POPOTO.GeoField')
from ..redis_db import POPOTO_REDIS_DB


class GeoField(Field):
    """
    A field that stores geospatial coordinates and enables geospatial search
    required: latitude, longitude
    """
    Coordinates = namedtuple('Coordinates', 'latitude longitude')

    type: type = tuple
    latitude: float = None
    longitude: float = None
    default: tuple = Coordinates(None, None)
    null: bool = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        geofield_defaults = {
            'type': tuple,
            'latitude': None,
            'longitude': None,
            'null': True,
            'default': GeoField.Coordinates(None, None),
        }
        self.field_defaults.update(geofield_defaults)
        # set field options, let kwargs override
        for k, v in geofield_defaults.items():
            setattr(self, k, kwargs.get(k, v))

    def get_filter_query_params(self, field_name):
        return super().get_filter_query_params(field_name) + [
            f'{field_name}',
            f'{field_name}__isnull',
            f'{field_name}_latitude',
            f'{field_name}_longitude',
            f'{field_name}_radius',
            f'{field_name}_radius_unit',
        ]

    @classmethod
    def is_valid(cls, field, value, null_check=True, **kwargs) -> bool:
        if not super().is_valid(field, value, null_check):
            return False
        if value is None:
            return True
        if isinstance(value, GeoField.Coordinates):
            pass
        elif isinstance(value, tuple):
            value = GeoField.Coordinates(value[0], value[1])
        else:
            logger.error(f"GeoField MUST be type GeoField.Coordinates or tuple, NOT {type(value)}")
            return False
        if field.null and not any([value.latitude, value.longitude]):
            return True
        elif bool(value.latitude) != bool(value.longitude):
            logger.error(f"latitude is {value.latitude} and longitude is {value.longitude}")
            logger.error(f"BOTH latitude AND longitude MUST have a value or be None")
            return False
        elif null_check and not all([value.latitude, value.longitude]):
            logger.error(f"latitude is {value.latitude} and longitude is {value.longitude}")
            logger.error(f"BOTH latitude AND longitude MUST have a value")
            return False
        try:
            float(value.latitude), float(value.longitude)
        except ValueError as e:
            logger.error(e)
            return False
        except TypeError as e:
            logger.error(e)
            return False
        return True

    def format_value_pre_save(self, field_value):
        """
        format field_value before saving to db
        return corrected field_value
        assumes validation is already passed
        """
        if isinstance(field_value, GeoField.Coordinates):
            return field_value
        if isinstance(field_value, tuple):
            return GeoField.Coordinates(field_value[0], field_value[1])
        if self.null:
            return GeoField.Coordinates(None, None)
        return field_value

    @classmethod
    def get_geo_db_key(cls, model, field_name: str) -> DB_key:
        return cls.get_special_use_field_db_key(model, field_name)

    @classmethod
    def on_save(cls, model_instance: 'Model', field_name: str, field_value: 'GeoField.Coordinates', pipeline=None, **kwargs):
        geo_db_key = cls.get_geo_db_key(model_instance, field_name)
        geo_member = model_instance.db_key.redis_key
        if not field_value or not (field_value.longitude and field_value.latitude):
            if pipeline:
                return pipeline.zrem(geo_db_key.redis_key, geo_member)
            else:
                return POPOTO_REDIS_DB.zrem(geo_db_key.redis_key, geo_member)
        if pipeline:
            return pipeline.geoadd(geo_db_key.redis_key, field_value.longitude, field_value.latitude, geo_member)
        else:
            return POPOTO_REDIS_DB.geoadd(geo_db_key.redis_key, field_value.longitude, field_value.latitude, geo_member)

    @classmethod
    def on_delete(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None, **kwargs):
        geo_db_key = cls.get_geo_db_key(model_instance, field_name)
        geo_member = model_instance.db_key.redis_key
        if pipeline:
            return pipeline.zrem(geo_db_key.redis_key, geo_member)
        else:
            return POPOTO_REDIS_DB.zrem(geo_db_key.redis_key, geo_member)

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """
        field = model._meta.fields[field_name]
        geo_db_key = cls.get_geo_db_key(model, field_name)
        coordinates = GeoField.Coordinates(None, None)
        member, radius, unit = None, 1, 'm'
        for query_param, query_value in query_params.items():

            if query_param == f'{field_name}':
                if isinstance(query_value, GeoField.Coordinates):
                    coordinates = query_value
                elif not isinstance(query_value, tuple) or not len(query_value) == 2:
                    from ..models.query import QueryException
                    raise QueryException(f"{query_param} must be assigned a tuple = (latitude, longitude)")
                else:
                    coordinates = GeoField.Coordinates(query_value[0], query_value[1])

            elif query_param.endswith('_latitude'):
                coordinates = GeoField.Coordinates(query_value, coordinates.longitude)
            elif query_param.endswith('_longitude'):
                coordinates = GeoField.Coordinates(coordinates.latitude, query_value)

            elif '_member' in query_param:
                if not isinstance(query_value, model):
                    from ..models.query import QueryException
                    raise QueryException(f"{query_param} must be assigned a tuple = (latitude, longitude)")
                member = query_value

            elif query_param.endswith('_radius'):
                radius = query_value

            elif query_param.endswith('_radius_unit'):
                if query_value not in ['m', 'km', 'ft', 'mi']:
                    from ..models.query import QueryException
                    raise QueryException(f"{query_param} must be one of m|km|ft|mi ")
                unit = query_value

        if member:
            redis_db_keys_list = POPOTO_REDIS_DB.georadiusbymember(
                geo_db_key.redis_key, member=member.db_key.redis_key,
                radius=radius, unit=unit  # , withdist=True, sort='asc'
            )

        elif coordinates.latitude and coordinates.longitude:
            # logger.debug(f"geo query on {dict(model=model._meta.db_class_key, longitude=coordinates.longitude, latitude=coordinates.latitude, radius=radius, unit=unit)}")
            redis_db_keys_list = POPOTO_REDIS_DB.georadius(
                geo_db_key.redis_key,
                longitude=coordinates.longitude, latitude=coordinates.latitude,
                radius=radius, unit=unit  # , withdist=True, sort='asc'
            )
            # logger.debug(f"geo query returned {redis_db_keys_list}")
        else:
            from ..models.query import QueryException
            raise QueryException(f"missing one or more required parameters. "
                                 f"geofilter requires either coordinates or instance of the same model")

        return set(redis_db_keys_list)



import logging
from src.popoto.exceptions import FinanceException
from ...finance import TickerStorage

#  timestamp_is_near_5min, get_nearest_5min_timestamp
# default_volume_indexes, derived_volume_indexes
# VOLUME_INDEXES

logger = logging.getLogger(__name__)


class VolumeException(FinanceException):
    pass


class VolumeStorage(TickerStorage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = kwargs.get('index', "close_volume")
        self.value = kwargs.get('value')
        self.db_key_suffix = f':{self.index}'

    def save(self, *args, **kwargs):

        # meets basic requirements for saving
        if not all([self.ticker, self.publisher,
                   self.index, self.value,
                   self.unix_timestamp]):
            logger.error("incomplete information, cannot save \n" + str(self.__dict__))
            raise VolumeException("save error, missing data")

        if not self.force_save:
            if not self.index in VOLUME_INDEXES:
                logger.error("volume index not in approved list, raising exception...")
                raise VolumeException("unknown index")

        if self.unix_timestamp % 3600 != 0:
            raise VolumeException("price timestamp should be % 3600")

        self.db_key_suffix = self.db_key_suffix or f':{self.index}'
        return super().save(*args, **kwargs)

    @classmethod
    def query(cls, *args, **kwargs):

        if kwargs.get("periods_key", None):
            raise PriceException("periods_key is not usable in PriceStorage query")

        key_suffix = kwargs.get("key_suffix", "")
        index = kwargs.get("index", "close_price")
        kwargs["key_suffix"] = f'{index}' + (f':{key_suffix}' if key_suffix else "")

        results_dict = super().query(*args, **kwargs)

        results_dict['index'] = index
        return results_dict



import logging
from decimal import Decimal
import datetime
import typing
import redis

from ..models.db_key import DB_key
from ..models.query import QueryException
from ..redis_db import POPOTO_REDIS_DB
logger = logging.getLogger('POPOTO.SortedFieldMixin')


class SortedFieldMixin:
    """
        The SortedField enables fast queries for ordering or filter by value range.
        Examples:
            Toys.query.filter(price__lte=4.99)
            DairyProduct.query.filter(best_before_date__gte=datetime.now())
        Requirements:
            Must be numeric type (int, float, Decimal, date, datetime)
            Null values not allowed. Can set a default.
    """
    type: type = float
    null: bool = False
    default = ""
    partition_on = tuple()

    def __init__(self, **kwargs):
        super().__init__()
        sortedfield_defaults = {
            'type': float,
            'null': False,
            'default': None,  # cannot set a default for datetime, so no type gets a default
            'partition_on': tuple(),
        }
        self.field_defaults.update(sortedfield_defaults)

        # set field_options, let kwargs override
        for k, v in sortedfield_defaults.items():
            setattr(self, k, kwargs.get(k, v))

        if isinstance(self.partition_on, str):
            self.partition_on = tuple((self.partition_on,))

        elif self.partition_on and not isinstance(self.partition_on, tuple):
            from ..models.base import ModelException
            raise ModelException("partition_on must be str or tuple of str field names")

        # todo: move this to field init validation
        if self.null is not False:
            from ..models.base import ModelException
            raise ModelException("SortedField cannot be null")
            # todo: allow null in SortedField. null removes instance from SortedSet
            # todo: if allow null, how to filter by null value? use extra index set just for nulls?

    def get_filter_query_params(self, field_name):
        return super().get_filter_query_params(field_name) + [
            f'{field_name}',
            f'{field_name}__gt',
            f'{field_name}__gte',
            f'{field_name}__lt',
            f'{field_name}__lte',
            # f'{field_name}__min',
            # f'{field_name}__max',
            # f'{field_name}__range',  # todo: like https://docs.djangoproject.com/en/3.2/ref/models/querysets/#range
            # f'{field_name}__isnull',  # todo: see todo in __init__
        ] + [
            f'{keyfield_name}' for keyfield_name in self.partition_on
        ]

    @classmethod
    def is_valid(cls, field, value, null_check=True, **kwargs) -> bool:
        if not super().is_valid(field, value, null_check):
            return
        if value and not isinstance(value, field.type):
            return False
        return True

    def format_value_pre_save(self, field_value):
        if self.type in [int, float, datetime.datetime, datetime.date, datetime.time]:
            return field_value
        else:
            return float(field_value)

    @classmethod
    def convert_to_numeric(cls, field, field_value):
        if field.type in [int, float]:
            return field_value
        elif field.type is Decimal:
            return float(field_value)
        elif field.type is datetime.date:
            return field_value.toordinal()
        elif field.type is datetime.datetime:
            return field_value.timestamp()
        elif field.type is datetime.time:
            return field_value.timestamp()
        else:
            raise ValueError("SortedField received non-numeric value.")

    @classmethod
    def get_sortedset_db_key(cls, model, field_name) -> DB_key:
        return cls.get_special_use_field_db_key(model, field_name)

    @classmethod
    def get_partitioned_sortedset_db_key(cls, model_instance, field_name) -> DB_key:
        sortedset_db_key = cls.get_sortedset_db_key(model_instance, field_name)
        # use field names and query values partition_on fields to extend sortedset_db_key
        for partition_field_name in model_instance._meta.fields[field_name].partition_on:
            try:
                sortedset_db_key.append(str(getattr(model_instance, partition_field_name)))
            except KeyError:
                raise QueryException(f"{field_name} filter requires partition_on field values")
        return sortedset_db_key

    @classmethod
    def on_save(cls, model_instance: 'Model', field_name: str, field_value: typing.Union[int, float],
                pipeline: redis.client.Pipeline = None, **kwargs):
        sortedset_db_key = cls.get_partitioned_sortedset_db_key(model_instance, field_name)

        sortedset_member = model_instance.db_key.redis_key
        sortedset_score = cls.convert_to_numeric(model_instance._meta.fields[field_name], field_value)

        if isinstance(pipeline, redis.client.Pipeline):
            return pipeline.zadd(sortedset_db_key.redis_key, {sortedset_member: sortedset_score})
        else:
            return POPOTO_REDIS_DB.zadd(sortedset_db_key.redis_key, {sortedset_member: sortedset_score})

    @classmethod
    def on_delete(cls, model_instance: 'Model', field_name: str, field_value, pipeline: redis.client.Pipeline = None,
                  **kwargs):
        sortedset_db_key = cls.get_partitioned_sortedset_db_key(model_instance, field_name)
        sortedset_member = model_instance.db_key.redis_key
        if pipeline:
            return pipeline.zrem(sortedset_db_key.redis_key, sortedset_member)
        else:
            return POPOTO_REDIS_DB.zrem(sortedset_db_key.redis_key, sortedset_member)

    @classmethod
    def filter_query(cls, model_class: 'Model', field_name: str, **query_params) -> set:
        """
        :param model_class: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """
        value_range = {'min': '-inf', 'max': '+inf'}

        for query_param, query_value in query_params.items():
            if field_name not in query_param:
                continue

            numeric_value = cls.convert_to_numeric(model_class._meta.fields[field_name], query_value)
            if '__gt' in query_param:
                inclusive = query_param.split('__gt')[1]
                value_range['min'] = f"{'' if inclusive == 'e' else '('}{numeric_value}"
            elif '__lt' in query_param:
                inclusive = query_param.split('__lt')[1]
                value_range['max'] = f"{'' if inclusive == 'e' else '('}{numeric_value}"
            else:
                raise QueryException(f"Query filters provided are not compatible with this field {field_name}")

        sortedset_db_key = cls.get_sortedset_db_key(model_class, field_name)

        # use field names and query values partition_on fields to extend sortedset_db_key
        for field_name in model_class._meta.fields[field_name].partition_on:
            try:
                sortedset_db_key += f":{str(query_params[field_name])}"
            except KeyError:
                raise QueryException(f"{field_name} filter requires partition_on field values")

        redis_db_keys_list = POPOTO_REDIS_DB.zrangebyscore(
            sortedset_db_key.redis_key, value_range['min'], value_range['max']
        )
        # redis_db_keys_list = POPOTO_REDIS_DB.zrange(
        #     sortedset_db_key.redis_key, value_range['min'], value_range['max'],
        #     desc=False, withscores=False,
        #     byscore=True, offset=None, num=None
        # )
        return set(redis_db_keys_list)



import numpy as np
from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS
from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH, BEARISH
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class RsiStorage(IndicatorStorage):

    def get_rsi_strength(self) -> int:
        rsi = int(self.value)
        if rsi is None or rsi <= 0.0 or rsi >= 100.0:
            return None

        assert (rsi>0.0) & (rsi<100.0), '>>> ERROR: RSI has extreme value of 0 or 100, highly unlikely'

        logger.debug(f"RSI={rsi}")

        rsi_strength = 0
        if rsi >= 80:
            rsi_strength = -3  # Extremely overbought
        elif rsi >= 75:
            rsi_strength = -2  # very overbought
        elif rsi >= 70:
            rsi_strength = -1  # overbought
        elif rsi <= 20:
            rsi_strength = 3  # Extremely oversold
        elif rsi <= 25:
            rsi_strength = 2   # very oversold
        elif rsi <= 30:
            rsi_strength = 1  # oversold
        return rsi_strength

    def produce_signal(self):

        rsi_strength = self.get_rsi_strength()
        if rsi_strength != 0:
            self.send_signal(
                trend=(BULLISH if rsi_strength > 0 else BEARISH),
                strength_value = int(np.abs(rsi_strength)), # should be 1,2,or3
                strength_max = int(3),
            )


class RsiSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if str(self.index) is not "close_price":
            logger.debug(f'index {self.index} is not close_price ...ignoring...')
            return

        new_rsi_storage = RsiStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        for horizon in HORIZONS:
            periods = horizon*14

            close_value_np_array = new_rsi_storage.get_denoted_price_array("close_price", periods)

            rsi_value = talib.RSI(close_value_np_array, timeperiod=periods)[-1]
            # logger.debug(f'savingRSI value {rsi_value} for {self.ticker} on {horizon*14} periods')

            new_rsi_storage.periods = horizon
            new_rsi_storage.value = float(rsi_value)
            new_rsi_storage.save()



from abc import ABC
import logging

import redis

from ..redis_db import POPOTO_REDIS_DB
import msgpack

logger = logging.getLogger('POPOTO-publisher')


class PublisherException(Exception):
    pass


class Publisher(ABC):
    _channel_name: str = ""
    _publish_data: dict = {}

    def __init__(self, *args, **kwargs):
        self._channel_name = kwargs.get('channel_name', self.__class__.__name__)
        super().__init__(*args, **kwargs)

    @property
    def channel_name(self):
        return self._channel_name

    @channel_name.setter
    def channel_name(self, value):
        self._channel_name = value

    def publish(self, data: dict = None, channel_name: str = None, pipeline: redis.client.Pipeline = None):
        import msgpack_numpy as m
        m.patch()
        # logger.debug(f"publish to {channel_name}: {publish_data}")
        channel_name = channel_name or self._channel_name
        self._publish_data = data or self._publish_data
        if self._publish_data is None or self._publish_data is {}:
            return None
        elif not channel_name:
            raise PublisherException("missing channel to publish to")

        if pipeline:
            return pipeline.publish(self._channel_name, msgpack.packb(self._publish_data))
        else:
            subscriber_count = POPOTO_REDIS_DB.publish(channel_name, msgpack.packb(self._publish_data))
            logger.debug(f"published data to `{channel_name}`, {subscriber_count} subscribers")
            return subscriber_count



import json
from abc import ABC
import logging
from json import JSONDecodeError
from src.popoto.exceptions import FinanceException
from ...redis_db import get_REDIS_DB

logger = logging.getLogger(__name__)


class SubscriberException(FinanceException):
    pass


class TickerSubscriber(ABC):
    class_describer = "ticker_subscriber"
    classes_subscribing_to = [
        # ...
    ]

    def __init__(self):
        self.database = get_REDIS_DB()
        self.pubsub = self.database.pubsub()
        logger.info(f'New pubsub for {self.__class__.__name__}')
        for s_class in self.classes_subscribing_to:
            self.pubsub.subscribe(s_class.__name__)
            logger.info(f'{self.__class__.__name__} subscribed to '
                        f'{s_class.__name__} channel')

    def __call__(self):
        data_event = self.pubsub.get_message()
        if not data_event:
            return
        if not data_event.get('type') == 'message':
            return

        # logger.debug(f'got message: {data_event}')

        # data_event = {
        #   'type': 'message',
        #   'pattern': None,
        #   'channel': b'PriceStorage',
        #   'data': b'{
        #       "key": f'{self.ticker}:{self.publisher}:PriceStorage:{periods}',
        #       "name": "9545225909:1533883300",
        #       "score": "1533883300"
        #   }'
        # }

        try:
            channel_name = data_event.get('channel').decode("utf-8")
            event_data = json.loads(data_event.get('data').decode("utf-8"))
            # logger.debug(f'handling event in {self.__class__.__name__}')
            self.pre_handle(channel_name, event_data)
            self.handle(channel_name, event_data)
        except KeyError as e:
            logger.warning(f'unexpected format: {data_event} ' + str(e))
            pass  # message not in expected format, just ignore
        except JSONDecodeError:
            logger.warning(f'unexpected data format: {data_event["data"]}')
            pass  # message not in expected format, just ignore
        except Exception as e:
            raise SubscriberException(f'Error calling {self.__class__.__name__}: ' + str(e))

    def pre_handle(self, channel, data, *args, **kwargs):
        pass

    def handle(self, channel, data, *args, **kwargs):
        """
        overwrite me with some logic
        :return: None
        """
        logger.warning(f'NEW MESSAGE for '
                       f'{self.__class__.__name__} subscribed to '
                       f'{channel} channel '
                       f'BUT HANDLER NOT DEFINED! '
                       f'... message/event discarded')
        pass


def get_nearest_1day_timestamp(timestamp) -> int:
    return ((int(timestamp) + 21600) // 86400) * 86400


def get_nearest_1day_score(score) -> int:
    return ((int(score) + 72) // 288) * 288


def timestamp_is_near_1hr(timestamp) -> bool:
    # close to a 1 hr minute period mark? (+ or - 45 seconds)
    seconds_from_1hr = (int(timestamp) + 150) % 3600
    return bool(seconds_from_1hr < 300)


def score_is_near_1hr(score) -> bool:
    return score % 12 < 3 or score % 12 > 9


def get_nearest_1hr_timestamp(timestamp) -> int:
    return ((int(timestamp) + 150) // 3600) * 3600


def get_nearest_1hr_score(score) -> int:
    return ((int(score) + 3) // 12) * 12


def timestamp_is_near_5min(timestamp) -> bool:
    # close to a five minute period mark? (+ or - 45 seconds)
    seconds_from_five_min = (int(timestamp) + 45) % 300
    return bool(seconds_from_five_min < 90)


def score_is_near_5min(score) -> bool:
    return bool(round(score) - 45 / 300 < score < round(score) + 45 / 300)


def get_nearest_5min_timestamp(timestamp) -> int:
    return ((int(timestamp) + 45) // 300) * 300


def get_nearest_5min_score(score) -> int:
    return int(round(score))



from models import TickerStorage, PriceStorage, VolumeStorage, PriceVolumeHistoryStorage, OLHCVModel, AssetModel, IndicatorStorage, PortfolioStorage
from subscribers import TickerSubscriber, IndicatorSubscriber



import datetime
from collections import namedtuple
from decimal import Decimal

import msgpack

from ..exceptions import ModelException
from ..fields.geo_field import GeoField
from ..redis_db import ENCODING

EncoderDecoder = namedtuple("EncoderDecoder", "key, encoder, decoder")

TYPE_ENCODER_DECODERS = {
    Decimal: EncoderDecoder(
        key='__Decimal__',
        encoder=lambda obj: {'__Decimal__': True, 'as_encodable': str(obj)},
        decoder=lambda obj: Decimal(obj['as_encodable'])
    ),
    tuple: EncoderDecoder(
        key='__tuple__',
        encoder=lambda obj: {'__tuple__': True, 'as_encodable': list(obj)},
        decoder=lambda obj: tuple(obj['as_encodable'])
    ),
    set: EncoderDecoder(
        key='__set__',
        encoder=lambda obj: {'__set__': True, 'as_encodable': list(obj)},
        decoder=lambda obj: set(obj['as_encodable'])
    ),
    datetime.datetime: EncoderDecoder(
        key='__datetime__',
        encoder=lambda obj: {'__datetime__': True, 'as_encodable': obj.strftime("%Y%m%dT%H:%M:%S.%f")},
        decoder=lambda obj: datetime.datetime.strptime(obj['as_encodable'], "%Y%m%dT%H:%M:%S.%f")
    ),
    datetime.date: EncoderDecoder(
        key='__date__',
        encoder=lambda obj: {'__date__': True, 'as_encodable': obj.strftime("%Y%m%d")},
        decoder=lambda obj: datetime.datetime.strptime(obj['as_encodable'], "%Y%m%d").date()
    ),
    datetime.time: EncoderDecoder(
        key='__time__',
        encoder=lambda obj: {'__time__': True, 'as_encodable': obj.strftime("%H:%M:%S.%f")},
        decoder=lambda obj: datetime.datetime.strptime(obj['as_encodable'], "%H:%M:%S.%f").time()
    ),
}
TYPE_DECODER_KEYSTRINGS = [ed.key for ed in TYPE_ENCODER_DECODERS.values()]


def decode_custom_types(obj):
    if isinstance(obj, dict) and 'as_encodable' in obj:
        for encoder_decoder in TYPE_ENCODER_DECODERS.values():
            if encoder_decoder.key in obj:
                return encoder_decoder.decoder(obj)
    return obj


def encode_popoto_model_obj(obj: 'Model') -> dict:
    import msgpack_numpy as m
    m.patch()

    encoded_hashmap = dict()
    for field_name, field in obj._meta.fields.items():
        value = getattr(obj, field_name)

        # use db_key string for relationships
        from ..fields.relationship import Relationship
        if value and isinstance(field, Relationship):
            if not isinstance(value, field.model):
                raise ModelException(f"Relationship field requires {field.model} model instance. got {value} instead")
            encoded_value = msgpack.packb(value.db_key.redis_key)
            # todo: refactor to store db_key list, not redis_key

        elif value and field.type in TYPE_ENCODER_DECODERS.keys():
            encoded_value = msgpack.packb(TYPE_ENCODER_DECODERS[field.type].encoder(value))
        else:
            encoded_value = msgpack.packb(value)

        encoded_hashmap[str(field_name).encode(ENCODING)] = encoded_value

    return encoded_hashmap


def decode_popoto_model_hashmap(model_class: 'Model', redis_hash: dict) -> 'Model':
    if len(redis_hash):
        return model_class(**{
            key_b.decode(ENCODING): decode_custom_types(msgpack.unpackb(value_b))
            for key_b, value_b in redis_hash.items()
        })
    else:
        return None



from settings import LOAD_TALIB
if LOAD_TALIB:
    import math, talib

from apps.TA.storages.abstract.indicator import IndicatorStorage
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class MacdStorage(IndicatorStorage):

    class_periods_list = [26]
    requisite_pv_indexes = ["close_price"]

    def compute_value_with_requisite_indexes(self, requisite_pv_index_arrrays: dict, periods: int = 0) -> str:
        """
        with cls.requisite_pv_indexes set

        :param index_value_arrrays: a dict with keys matching requisite+pv_indexes and values from self.get_denoted_price_array()
        :param periods: number of periods to compute value for
        :return:
        """
        periods = periods or self.periods
        fastperiod = periods*12/26
        slowperiod = periods*26/26
        signalperiod = periods*9/26

        macd_value, macdsignal, macdhist = talib.MACD(
            requisite_pv_index_arrrays["close_price"],
            fastperiod=fastperiod, slowperiod=slowperiod, signalperiod=signalperiod
        )

        logger.debug(f"Macd computed: {macd_value[-1]}:{macdsignal[-1]}:{macdhist[-1]}")

        if math.isnan(sum([macd_value[-1], macdsignal[-1], macdhist[-1]])): return ""

        return f"{macd_value[-1]}:{macdsignal[-1]}:{macdhist[-1]}"

    def produce_signal(self):
        pass

class MacdSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [PriceStorage]
    storage_class = MacdStorage



import logging
from datetime import datetime

from src.popoto.finance import PriceStorage, VolumeStorage, OLHCVModel

# import pandas
# from polygon import RESTClient as PolygonAPI
# from settings import POLYGON_API_KEY
# get_nearest_1hr_timestamp
# PUBLISHERS, yahoo_index_dict

logger = logging.getLogger(__name__)


ASSETS = {
    "stocks": [
        "FB", "GOOG", "AMZN", "MSFT", "ABNB", "RIOT", "UFO",
    ],
    "crypto": [
        "BTC", "ETH", "LTC",
    ],
    "currencies": [
        "USD", "EUR", "GBP", "AUD", "THB", "CZK",
    ]
}

def get_asset_class(symbol):
    asset_class = "other"
    for asset_class_key in ASSETS.keys():
        if symbol in ASSETS[asset_class_key]:
            return asset_class_key
    return asset_class


class AssetException(Exception):
    pass


class AssetModel(OLHCVModel):

    def __init__(self, symbol: str, asset_class: str = None, timestamp: int = 0, days_range: int = 1, period="daily"):
        if asset_class in ASSETS.keys():
            if symbol not in ASSETS[asset_class]:
                raise AssetException("symbol not found in this asset class")

        self.asset_class = asset_class or get_asset_class(symbol)
        super().__init__(symbol=symbol, publisher=PUBLISHERS.get(self.asset_class, "yahoo"),
                         timestamp=timestamp, days_range=days_range, period=period)


    @classmethod
    def query(cls, symbol, publisher="", *args, **kwargs):
        from apps.portfolio.models import Asset
        assets = Asset.objects.filter(name__icontains=symbol)
        asset = assets.filter(name__icontains=publisher).first() if publisher else assets.first()
        if asset:
            return AssetModel(
                symbol=asset.symbol, asset_class=asset.asset_class, days_range=90
            ).get_dataframe()


    def renew(self, start_timestamp=None, end_timestamp=None):  # get up-to-date with newest data
        from ...redis_db import POPOTO_REDIS_DB
        end_timestamp = end_timestamp or self.timestamp
        one_day_seconds = 24 * 3600
        start_timestamp = start_timestamp or end_timestamp - (self.days_range * one_day_seconds)
        start_timestamp -= one_day_seconds
        throttle_timeout = 20


        if POPOTO_REDIS_DB.ttl(f"throttle:{PUBLISHERS[self.asset_class]}") > 0:
            return
        POPOTO_REDIS_DB.set(f"throttle:{PUBLISHERS[self.asset_class]}", "slow up bro", ex=throttle_timeout) #set throttle

        pipeline = POPOTO_REDIS_DB.pipeline()

        if PUBLISHERS[self.asset_class] == "polygon":
            with PolygonAPI(POLYGON_API_KEY) as polygon_client:
                cryptoTicker, multiplier, timespan = f"X:{self.symbol}USD", 1, 'day'
                to = datetime.fromtimestamp(end_timestamp).strftime('%Y-%m-%d')
                from_ = datetime.fromtimestamp(start_timestamp).strftime('%Y-%m-%d')

                endpoint = f"{polygon_client.url}/v2/aggs/ticker/{cryptoTicker}/range/{multiplier}/{timespan}/{from_}/{to}"
                polygon_client_response = polygon_client._session.get(endpoint, params={},
                                                                      timeout=polygon_client.timeout)
                logging.debug(f"polygon responded with status: {polygon_client_response.status_code}")

            if polygon_client_response.status_code == 200 and polygon_client_response.json()['results']:

                for aggregate_result in polygon_client_response.json()['results']:
                    for agg_key, index_name in {
                        'o': 'open_price',
                        'h': 'high_price',
                        'l': 'low_price',
                        'c': 'close_price',
                    }.items():
                        price_storage = PriceStorage(
                            ticker=f"{self.symbol}_USD",
                            publisher="polygon",
                            index=index_name,
                            timestamp=get_nearest_1hr_timestamp(float(aggregate_result['t'])/1000),
                            value=float(aggregate_result[agg_key])
                        )
                        pipeline = price_storage.save(pipeline=pipeline)

                    volume_storage = VolumeStorage(
                        ticker=f"{self.symbol}_USD",
                        publisher='polygon',
                        index='close_volume',
                        timestamp=get_nearest_1hr_timestamp(float(aggregate_result['t']) / 1000),
                        value = float(aggregate_result['v'])
                    )
                    pipeline = volume_storage.save(pipeline=pipeline)


        if PUBLISHERS[self.asset_class] == "yahoo":
            from yahoo_fin.stock_info import get_data
            self.yahoo_data = get_data(
                self.symbol,
                start_date=pandas.Timestamp.utcfromtimestamp(start_timestamp),
                end_date=pandas.Timestamp.utcfromtimestamp(end_timestamp)
            )
            for index in set(self.yahoo_data.keys()).intersection({'open', 'high', 'low', 'close'}):
                for timestamp, value in self.yahoo_data[index].items():
                    price_storage = PriceStorage(
                        ticker=f"{self.symbol}_USD",
                        publisher="yahoo",
                        index=yahoo_index_dict[index],
                        timestamp=int(timestamp.timestamp()),
                        value=float(value)
                    )
                    price_storage.save(pipeline=pipeline)
            for index in set(self.yahoo_data.keys()).intersection({'volume'}):
                for timestamp, value in self.yahoo_data[index].items():
                    volume_storage = VolumeStorage(
                        ticker=f"{self.symbol}_USD",
                        publisher="yahoo",
                        index=yahoo_index_dict[index],
                        timestamp=int(timestamp.timestamp()),
                        value=float(value)
                    )
                    volume_storage.save(pipeline=pipeline)

        pipeline.execute()
        self.dataframe = self.get_dataframe()
        return



import logging

import redis
from redisgraph import Graph
from .._archive.key_value import KeyValueModel
from ..redis_db import POPOTO_REDIS_DB
from redisgraph import Node as GraphNode
from redisgraph import Edge as GraphEdge
from ..exceptions import ModelException

logger = logging.getLogger(__name__)


class GraphNodeException(ModelException):
    pass


class GraphNodeModel(KeyValueModel):
    """
    stores things and their relationships in redis graph database
    otherwise acts as KeyValueModel
    recommend to uniquely identify the instance with a key prefix or suffix
    prefixes are for more specific categories of objects (eg. mammal:human:woman:Lisa )
    suffixes are for specific attributes (eg. Lisa:eye_color, Lisa:age, etc)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graph = Graph('graph', POPOTO_REDIS_DB)
        

    @property
    def graph_node(self):
        return GraphNode(
            alias=self._db_key,
            label=self.__class__.__name__,
            properties=self.value if isinstance(self.value, dict) else {'value': self.value}
        )

    @property
    def relationships(self, relationship_type: str, filters: dict = None):
        # params = {'purpose': "pleasure"}
        # query = """MATCH (p:person)-[v:visited {purpose:$purpose}]->(c:country)
        # 		   RETURN p.name, p.age, v.purpose, c.name"""

        self.query(f"""MATCH (from:{self.__class__.__name__})-[rel:{relationship_type}]->(to:{self.__class__.__name__})
                    RETURN from.db_key, rel.db_key, to.db_key """)


    def save(self, pipeline: redis.client.Pipeline = None, *args, **kwargs):
        super().save(pipeline, *args, **kwargs)
        self.graph.add_node(self.graph_node)
        self.graph.commit()


    def query(self, query_string: str, params: dict = None):
        result = self.graph.query(query_string, params)
        result_db_keys = [record[2] for record in result.result_set]
        return [self.__class__.get(db_key) for db_key in result_db_keys]


    def _set_relationship_to_graphnode(self, context: dict, graph_node: GraphNode) -> None:
        self.graph_edges.append(
            GraphEdge(self.graph_node, 'correlation', graph_node, properties=context)
        )

    def save(self):
        self.graph.add_node(self.graph_node)
        for edge in self.graph_edges:
            self.graph.add_edge(edge)
        self.graph.commit()


# class GraphRelationship:
#     def __init__(self):
#         self.__name__



from .fields.field import Field
from .fields.shortcuts import IntField, FloatField, DecimalField, StringField, BooleanField, ListField, DictField, SetField, TupleField, BytesField, DateField, DatetimeField, TimeField, KeyField, AutoKeyField, UniqueKeyField, SortedField, SortedKeyField
from .fields.geo_field import GeoField
from .fields.relationship import Relationship
from .models.base import Model, ModelBase
from .pubsub.publisher import Publisher
from .pubsub.subscriber import Subscriber

__all__ = [
    'Field',
    'IntField', 'FloatField', 'DecimalField', 'StringField',
    'BooleanField', 'BytesField',
    'ListField', 'DictField', 'SetField', 'TupleField',
    'DateField', 'DatetimeField', 'TimeField',
    'KeyField', 'AutoKeyField', 'UniqueKeyField',
    'SortedField', 'SortedKeyField', 'GeoField',
    'Relationship',
    'Model', 'ModelBase',
    'Publisher', 'Subscriber',
]



import logging
logger = logging.getLogger('POPOTO.KeyFieldMixin')


class UniqueFieldMixin:
    """
    UniqueKeyField() is equivalent to KeyField(unique=True)
    """
    unique: bool = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        uniquekeyfield_defaults = {
            'unique': True,
        }
        self.field_defaults.update(uniquekeyfield_defaults)
        # set field options, let kwargs override
        for k, v in uniquekeyfield_defaults.items():
            setattr(self, k, kwargs.get(k, v))

        if not kwargs.get('unique', True):
            from ..models.base import ModelException
            raise ModelException("UniqueKey field MUST be unique")





from abc import ABC
from redisgraph import Node as GraphNode
from redisgraph import Edge as GraphEdge
from redisgraph import Graph, Path


class AbstractNode(ABC):
    graph_node = GraphNode()
    graph_edges = []

    def __init__(self, name, *args, **kwargs):
        self.base_class = self.__class__.__name__
        self.graph_node = GraphNode(
            label=name,
            properties={'id': 'uuid', 'importance': 1, 'active': True}
        )

    def _set_relationship_to_graphnode(self, context: dict, graph_node: GraphNode) -> None:
        self.graph_edges.append(
            GraphEdge(self.graph_node, 'correlation', graph_node, properties=context)
        )

    def save(self):
        pass
        # redis_graph = Graph(self.base_class, redis_db)
        # redis_graph.add_node(self.graph_node)
        # for e in self.graph_edges:
        #     redis_graph.add_edge(e)
        # redis_graph.commit()


class GenericNode(AbstractNode):

    def __init__(self, name):
        self.parent_class = None
        super().__init__(self, name)



from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS
from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger

KAMA_LIST = [30,]


class KamaStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class KamaSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if self.index != 'close_price':
            logger.debug(f'index {self.index} is not close_price ...ignoring...')
            return

        new_kama_storage = KamaStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        periods_list = []
        for s in KAMA_LIST:
            periods_list.extend([h * s for h in HORIZONS])

        for periods in set(periods_list):

            close_value_np_array = new_kama_storage.get_denoted_price_array("close_price", periods)

            kama_value = talib.KAMA(close_value_np_array, timeperiod=periods)[-1]
            # logger.debug(f'savingKama value {kama_value}for {self.ticker} on {periods} periods')

            new_kama_storage.periods = periods
            new_kama_storage.value = float(kama_value)
            new_kama_storage.save()



# A DECORATOR FOR PYTHON THREADING
# http://docs.python.org/2/library/threading.html#thread-objects
# http://stackoverflow.com/questions/18420699/multithreading-for-python-django
from threading import Thread

def start_new_thread(function):
    """
    if running Django relational database transactions,
    include this at the end of your function
        ```
        from django.db import connection
        connection.close()
        ```
    """
    def decorator(*args, **kwargs):
        t = Thread(target = function, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()
    return decorator


def run_all_multithreaded(function_def, list_of_params):
    """
    :param function_def: the function to pass params to
    :param list_of_params: list of function params, for multiple-param functions, pass tuples
    :return:
    """

    from multiprocessing.dummy import Pool as ThreadPool

    # make the Pool of workers
    pool = ThreadPool(16)

    # open functions in their own threads
    # and compile all the results
    results = pool.map(function_def, list_of_params)

    # close the pool and wait for the work to finish
    pool.close()
    pool.join()

    return results



from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS
from apps.TA.storages.abstract.indicator import IndicatorStorage, BULLISH, BEARISH
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class WillrStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class WillrSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if str(self.index) is not "close_price":
            logger.debug(f'index {self.index} is not close_price ...ignoring...')
            return

        new_willr_storage = WillrStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        for horizon in HORIZONS:
            periods = horizon * 14

            high_value_np_array = new_willr_storage.get_denoted_price_array("high_price", periods)
            low_value_np_array = new_willr_storage.get_denoted_price_array("low_price", periods)
            close_value_np_array = new_willr_storage.get_denoted_price_array("close_price", periods)

            willr_value = talib.WILLR(high_value_np_array, low_value_np_array, close_value_np_array,
                                      timeperiod=horizon*14)[-1]
            # logger.debug(f'savingWillr value {willr_value} for {self.ticker} on {periods} periods')

            new_willr_storage.periods = periods
            new_willr_storage.value = float(willr_value)
            new_willr_storage.save()



from collections.abc import Iterable

from ..redis_db import POPOTO_REDIS_DB


class DB_key(list):
    def __init__(self, *key_partials):
        def flatten(yet_flat):
            if isinstance(yet_flat, Iterable) and not isinstance(yet_flat, (str, bytes)):
                for item in yet_flat:
                    yield from flatten(item)
            else:
                yield yet_flat

        super().__init__(flatten(key_partials))

    @classmethod
    def clean(cls, value: str, ignore_colons: bool = False) -> str:
        value = value.replace('/', '//')
        for char in "'?*^[]-":
            value = value.replace(char, f"/{char}")
        if not ignore_colons:
            value = value.replace(':', '_')
        return value

    def __str__(self):
        return ":".join([
            str(partial) if isinstance(partial, DB_key) else self.clean(str(partial))
            for partial in self
        ])

    @property
    def redis_key(self):
        return str(self)

    def exists(self):
        return True if POPOTO_REDIS_DB.exists(self.redis_key) > 0 else False

    def get_instance(self, model_class):
        redis_hash = POPOTO_REDIS_DB.hgetall(self.redis_key)
        from .encoding import decode_popoto_model_hashmap
        return decode_popoto_model_hashmap(model_class, redis_hash)



from ticker_subscriber import TickerSubscriber
from ...finance import PriceVolumeHistoryStorage

class VolumeSubscriber(TickerSubscriber):

    classes_subscribing_to = [
        PriceVolumeHistoryStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        # parse timestamp from data
        # f'{data_history.ticker}:{data_history.publisher}:{data_history.timestamp}'
        [ticker, publisher, timestamp] = data.split(":")

        if not timestamp_is_near_5min(timestamp):
            return

        # logger.debug("near to a 5 min time marker")
        timestamp = get_nearest_5min_timestamp(timestamp)

        volume = VolumeStorage(ticker=ticker, publisher=publisher, timestamp=timestamp)
        index_values = {}

        for index in default_volume_indexes:
            logger.debug(f'process volume for ticker: {ticker}')

            # example key = "XPM_BTC:poloniex:PriceVolumeHistoryStorage:close_price"
            sorted_set_key = f'{ticker}:{publisher}:PriceVolumeHistoryStorage:{index}'

            index_values[index] = [
                float(db_value.decode("utf-8").split(":")[0])
                for db_value
                in self.database.zrangebyscore(sorted_set_key, timestamp - 300, timestamp + 45) # todo: update to scores
            ]

            try:
                if not len(index_values[index]):
                    volume.value = None

                elif index == "close_volume":
                    volume.value = index_values["close_volume"][-1]


            except IndexError:
                pass  # couldn't find a useful value
            except ValueError:
                pass  # couldn't find a useful value
            else:
                if volume.value:
                    volume.index = index
                    volume.save()
                    # logger.info("saved new thing: " + volume.get_db_key())

            # all_values_set = (
            #         set(index_values["open_volume"])
            # )
            #
            # if not len(all_values_set):
            #     return
            #
            # for index in derived_volume_indexes:
            #     volume.value = None
            #     values_set = all_values_set.copy()
            #
            #     if index == "open_volume":
            #         volume.value = index_values["open_volume"][0]
            #
            #     elif index == "low_volume":
            #         volume.value = min(index_values["low_volume"])
            #     elif index == "high_volume":
            #         volume.value = max(index_values["high_volume"])
            #
            #     if volume.value:
            #         volume.index = index
            #         volume.save(publish=True)
